import unittest
import collections
from fontParts.base import FontPartsError


class TestGuideline(unittest.TestCase):
    def getGuideline_generic(self):
        guideline, _ = self.objectGenerator("guideline")
        guideline.x = 1
        guideline.y = 2
        guideline.angle = 90
        guideline.name = "Test Guideline"
        return guideline

    def getGuideline_fontGuideline(self):
        font, _ = self.objectGenerator("font")
        guideline = font.appendGuideline((1, 2), 90, "Test Guideline Font")
        return guideline

    def getGuideline_glyphGuideline(self):
        font, _ = self.objectGenerator("font")
        layer = font.newLayer("L")
        glyph = layer.newGlyph("X")
        guideline = glyph.appendGuideline((1, 2), 90, "Test Guideline Glyph")
        return guideline

    # ----
    # repr
    # ----

    def test_reprContents(self):
        guideline = self.getGuideline_generic()
        value = guideline._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    def test_reprContents_noGlyph(self):
        guideline, _ = self.objectGenerator("guideline")
        value = guideline._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    def test_reprContents_Layer(self):
        guideline = self.getGuideline_glyphGuideline()
        value = guideline._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    # --------
    # Attributes
    # --------

    # x

    def test_x_get_generic(self):
        guideline = self.getGuideline_generic()
        self.assertEqual(guideline.x, 1)

    def test_x_get_fontGuideline(self):
        guideline = self.getGuideline_fontGuideline()
        self.assertEqual(guideline.x, 1)

    def test_x_get_glyphGuideline(self):
        guideline = self.getGuideline_glyphGuideline()
        self.assertEqual(guideline.x, 1)

    def test_x_set_valid_zero_generic(self):
        guideline = self.getGuideline_generic()
        guideline.x = 0
        self.assertEqual(guideline.x, 0)

    def test_x_set_valid_zero_fontGuideline(self):
        guideline = self.getGuideline_fontGuideline()
        guideline.x = 0
        self.assertEqual(guideline.x, 0)

    def test_x_set_valid_zero_glyphGuideline(self):
        guideline = self.getGuideline_glyphGuideline()
        guideline.x = 0
        self.assertEqual(guideline.x, 0)

    def test_x_set_valid_positive(self):
        guideline = self.getGuideline_generic()
        guideline.x = 1
        self.assertEqual(guideline.x, 1)

    def test_x_set_valid_negative(self):
        guideline = self.getGuideline_generic()
        guideline.x = -1
        self.assertEqual(guideline.x, -1)

    def test_x_set_valid_positive_float(self):
        guideline = self.getGuideline_generic()
        guideline.x = 1.1
        self.assertEqual(guideline.x, 1.1)

    def test_x_set_valid_negative_float(self):
        guideline = self.getGuideline_generic()
        guideline.x = -1.1
        self.assertEqual(guideline.x, -1.1)

    def test_x_set_valid_None(self):
        guideline = self.getGuideline_generic()
        guideline.x = None
        self.assertEqual(guideline.x, 0)

    def test_x_set_invalid_string(self):
        guideline = self.getGuideline_generic()
        with self.assertRaises(TypeError):
            guideline.x = "ABC"

    # y

    def test_y_get_generic(self):
        guideline = self.getGuideline_generic()
        self.assertEqual(guideline.y, 2)

    def test_y_get_fontGuideline(self):
        guideline = self.getGuideline_fontGuideline()
        self.assertEqual(guideline.y, 2)

    def test_y_get_glyphGuideline(self):
        guideline = self.getGuideline_glyphGuideline()
        self.assertEqual(guideline.y, 2)

    def test_y_set_valid_zero_generic(self):
        guideline = self.getGuideline_generic()
        guideline.y = 0
        self.assertEqual(guideline.y, 0)

    def test_y_set_valid_zero_fontGuideline(self):
        guideline = self.getGuideline_fontGuideline()
        guideline.y = 0
        self.assertEqual(guideline.y, 0)

    def test_y_set_valid_zero_glyphGuideline(self):
        guideline = self.getGuideline_glyphGuideline()
        guideline.y = 0
        self.assertEqual(guideline.y, 0)

    def test_y_set_valid_positive(self):
        guideline = self.getGuideline_generic()
        guideline.y = 1
        self.assertEqual(guideline.y, 1)

    def test_y_set_valid_negative(self):
        guideline = self.getGuideline_generic()
        guideline.y = -1
        self.assertEqual(guideline.y, -1)

    def test_y_set_valid_positive_float(self):
        guideline = self.getGuideline_generic()
        guideline.y = 1.1
        self.assertEqual(guideline.y, 1.1)

    def test_y_set_valid_negative_float(self):
        guideline = self.getGuideline_generic()
        guideline.y = -1.1
        self.assertEqual(guideline.y, -1.1)

    def test_y_set_valid_None(self):
        guideline = self.getGuideline_generic()
        guideline.y = None
        self.assertEqual(guideline.y, 0)

    def test_y_set_invalid_string(self):
        guideline = self.getGuideline_generic()
        with self.assertRaises(TypeError):
            guideline.y = "ABC"

    # angle

    def test_angle_get_generic(self):
        guideline = self.getGuideline_generic()
        self.assertEqual(guideline.angle, 90)

    def test_angle_get_fontGuideline(self):
        guideline = self.getGuideline_fontGuideline()
        self.assertEqual(guideline.angle, 90)

    def test_angle_get_glyphGuideline(self):
        guideline = self.getGuideline_glyphGuideline()
        self.assertEqual(guideline.angle, 90)

    def test_angle_set_valid_zero_generic(self):
        guideline = self.getGuideline_generic()
        guideline.angle = 0
        self.assertEqual(guideline.angle, 0)

    def test_angle_set_valid_zero_fontGuideline(self):
        guideline = self.getGuideline_fontGuideline()
        guideline.angle = 0
        self.assertEqual(guideline.angle, 0)

    def test_angle_set_valid_zero_glyphGuideline(self):
        guideline = self.getGuideline_glyphGuideline()
        guideline.angle = 0
        self.assertEqual(guideline.angle, 0)

    def test_angle_set_valid_positive(self):
        guideline = self.getGuideline_generic()
        guideline.angle = 10
        self.assertEqual(guideline.angle, 10)

    def test_angle_set_valid_negative(self):
        guideline = self.getGuideline_generic()
        guideline.angle = -10
        self.assertEqual(guideline.angle, 350)

    def test_angle_set_valid_positive_float(self):
        guideline = self.getGuideline_generic()
        guideline.angle = 10.1
        self.assertEqual(guideline.angle, 10.1)

    def test_angle_set_valid_negative_float(self):
        guideline = self.getGuideline_generic()
        guideline.angle = -10.1
        self.assertEqual(guideline.angle, 349.9)

    def test_angle_set_valid_positive_edge(self):
        guideline = self.getGuideline_generic()
        guideline.angle = 360
        self.assertEqual(guideline.angle, 360)

    def test_angle_set_valid_negative_edge(self):
        guideline = self.getGuideline_generic()
        guideline.angle = -360
        self.assertEqual(guideline.angle, 0)

    def test_angle_set_valid_None(self):
        guideline = self.getGuideline_generic()
        guideline.angle = None
        self.assertEqual(guideline.angle, 0)

    def test_angle_set_invalid_positive_edge(self):
        guideline = self.getGuideline_generic()
        with self.assertRaises(ValueError):
            guideline.angle = 361

    def test_angle_set_invalid_negative_edge(self):
        guideline = self.getGuideline_generic()
        with self.assertRaises(ValueError):
            guideline.angle = -361

    def test_angle_set_invalid_string(self):
        guideline = self.getGuideline_generic()
        with self.assertRaises(TypeError):
            guideline.angle = "ABC"

    def test_angle_set_valid_none_x0_y0(self):
        guideline = self.getGuideline_generic()
        guideline.x = 0
        guideline.y = 0
        guideline.angle = None
        self.assertEqual(guideline.angle, 0)

    def test_angle_set_valid_none_x1_y0(self):
        guideline = self.getGuideline_generic()
        guideline.x = 1
        guideline.y = 0
        guideline.angle = None
        self.assertEqual(guideline.angle, 90)

    def test_angle_set_valid_none_x0_y1(self):
        guideline = self.getGuideline_generic()
        guideline.x = 0
        guideline.y = 1
        guideline.angle = None
        self.assertEqual(guideline.angle, 0)

    def test_angle_set_valid_none_x1_y1(self):
        guideline = self.getGuideline_generic()
        guideline.x = 1
        guideline.y = 1
        guideline.angle = None
        self.assertEqual(guideline.angle, 0)

    # index

    def getGuideline_index(self):
        glyph, _ = self.objectGenerator("glyph")
        glyph.appendGuideline((0, 0), 90, "guideline 0")
        glyph.appendGuideline((0, 0), 90, "guideline 1")
        glyph.appendGuideline((0, 0), 90, "guideline 2")
        return glyph

    def test_get_index_noParent(self):
        guideline, _ = self.objectGenerator("guideline")
        self.assertIsNone(guideline.index)

    def test_get_index(self):
        glyph = self.getGuideline_index()
        for i, guideline in enumerate(glyph.guidelines):
            self.assertEqual(guideline.index, i)

    def test_set_index_noParent(self):
        guideline, _ = self.objectGenerator("guideline")
        with self.assertRaises(FontPartsError):
            guideline.index = 1

    def test_set_index_positive(self):
        glyph = self.getGuideline_index()
        guideline = glyph.guidelines[0]
        with self.assertRaises(FontPartsError):
            guideline.index = 2

    def test_set_index_negative(self):
        glyph = self.getGuideline_index()
        guideline = glyph.guidelines[1]
        with self.assertRaises(FontPartsError):
            guideline.index = -1

    # name

    def test_name_get_none(self):
        guideline, _ = self.objectGenerator("guideline")
        self.assertIsNone(guideline.name)

    def test_name_set_valid(self):
        guideline = self.getGuideline_generic()
        guideline.name = "foo"
        self.assertEqual(guideline.name, "foo")

    def test_name_set_none(self):
        guideline = self.getGuideline_generic()
        guideline.name = None
        self.assertIsNone(guideline.name)

    def test_name_set_invalid(self):
        guideline = self.getGuideline_generic()
        with self.assertRaises(TypeError):
            guideline.name = 123

    # color

    def test_color_get_none(self):
        guideline = self.getGuideline_generic()
        self.assertIsNone(guideline.color)

    def test_color_set_valid_max(self):
        guideline = self.getGuideline_generic()
        guideline.color = (1, 1, 1, 1)
        self.assertEqual(guideline.color, (1, 1, 1, 1))

    def test_color_set_valid_min(self):
        guideline = self.getGuideline_generic()
        guideline.color = (0, 0, 0, 0)
        self.assertEqual(guideline.color, (0, 0, 0, 0))

    def test_color_set_valid_decimal(self):
        guideline = self.getGuideline_generic()
        guideline.color = (0.1, 0.2, 0.3, 0.4)
        self.assertEqual(guideline.color, (0.1, 0.2, 0.3, 0.4))

    def test_color_set_none(self):
        guideline = self.getGuideline_generic()
        guideline.color = None
        self.assertIsNone(guideline.color)

    def test_color_set_invalid_over_max(self):
        guideline = self.getGuideline_generic()
        with self.assertRaises(ValueError):
            guideline.color = (1.1, 0.2, 0.3, 0.4)

    def test_color_set_invalid_uner_min(self):
        guideline = self.getGuideline_generic()
        with self.assertRaises(ValueError):
            guideline.color = (-0.1, 0.2, 0.3, 0.4)

    def test_color_set_invalid_too_few(self):
        guideline = self.getGuideline_generic()
        with self.assertRaises(ValueError):
            guideline.color = (0.1, 0.2, 0.3)

    def test_color_set_invalid_string(self):
        guideline = self.getGuideline_generic()
        with self.assertRaises(TypeError):
            guideline.color = "0.1,0.2,0.3,0.4"

    def test_color_set_invalid_int(self):
        guideline = self.getGuideline_generic()
        with self.assertRaises(TypeError):
            guideline.color = 123

    # identifier

    def test_identifier_get_none(self):
        guideline = self.getGuideline_generic()
        self.assertIsNone(guideline.identifier)

    def test_identifier_generated_type(self):
        guideline = self.getGuideline_generic()
        guideline.getIdentifier()
        self.assertIsInstance(guideline.identifier, str)

    def test_identifier_consistency(self):
        guideline = self.getGuideline_generic()
        guideline.getIdentifier()
        # get: twice to test consistency
        self.assertEqual(guideline.identifier, guideline.identifier)

    def test_identifier_cannot_set(self):
        # identifier is a read-only property
        guideline = self.getGuideline_generic()
        with self.assertRaises(FontPartsError):
            guideline.identifier = "ABC"

    def test_identifier_force_set(self):
        identifier = "ABC"
        guideline = self.getGuideline_generic()
        guideline._setIdentifier(identifier)
        self.assertEqual(guideline.identifier, identifier)

    # -------
    # Methods
    # -------

    def getGuideline_copy(self):
        guideline = self.getGuideline_generic()
        guideline.name = "foo"
        guideline.color = (0.1, 0.2, 0.3, 0.4)
        return guideline

    # copy

    def test_copy_seperate_objects(self):
        guideline = self.getGuideline_copy()
        copied = guideline.copy()
        self.assertIsNot(guideline, copied)

    def test_copy_same_name(self):
        guideline = self.getGuideline_copy()
        copied = guideline.copy()
        self.assertEqual(guideline.name, copied.name)

    def test_copy_same_color(self):
        guideline = self.getGuideline_copy()
        copied = guideline.copy()
        self.assertEqual(guideline.color, copied.color)

    def test_copy_same_identifier(self):
        guideline = self.getGuideline_copy()
        copied = guideline.copy()
        self.assertEqual(guideline.identifier, copied.identifier)

    def test_copy_generated_identifier_different(self):
        guideline = self.getGuideline_copy()
        copied = guideline.copy()
        guideline.getIdentifier()
        copied.getIdentifier()
        self.assertNotEqual(guideline.identifier, copied.identifier)

    def test_copy_same_x(self):
        guideline = self.getGuideline_copy()
        copied = guideline.copy()
        self.assertEqual(guideline.x, copied.x)

    def test_copy_same_y(self):
        guideline = self.getGuideline_copy()
        copied = guideline.copy()
        self.assertEqual(guideline.y, copied.y)

    def test_copy_same_angle(self):
        guideline = self.getGuideline_copy()
        copied = guideline.copy()
        self.assertEqual(guideline.angle, copied.angle)

    # transform

    def getGuideline_transform(self):
        guideline = self.getGuideline_generic()
        guideline.angle = 45.0
        return guideline

    def test_transformBy_valid_no_origin(self):
        guideline = self.getGuideline_transform()
        guideline.transformBy((2, 0, 0, 3, -3, 2))
        self.assertEqual(guideline.x, -1)
        self.assertEqual(guideline.y, 8)
        self.assertAlmostEqual(guideline.angle, 56.310, places=3)

    def test_transformBy_valid_origin(self):
        guideline = self.getGuideline_transform()
        guideline.transformBy((2, 0, 0, 2, 0, 0), origin=(1, 2))
        self.assertEqual(guideline.x, 1)
        self.assertEqual(guideline.y, 2)
        self.assertAlmostEqual(guideline.angle, 45.000, places=3)

    def test_transformBy_invalid_one_string_value(self):
        guideline = self.getGuideline_transform()
        with self.assertRaises(TypeError):
            guideline.transformBy((1, 0, 0, 1, 0, "0"))

    def test_transformBy_invalid_all_string_values(self):
        guideline = self.getGuideline_transform()
        with self.assertRaises(TypeError):
            guideline.transformBy("1, 0, 0, 1, 0, 0")

    def test_transformBy_invalid_int_value(self):
        guideline = self.getGuideline_transform()
        with self.assertRaises(TypeError):
            guideline.transformBy(123)

    # moveBy

    def test_moveBy_valid(self):
        guideline = self.getGuideline_transform()
        guideline.moveBy((-1, 2))
        self.assertEqual(guideline.x, 0)
        self.assertEqual(guideline.y, 4)
        self.assertAlmostEqual(guideline.angle, 45.000, places=3)

    def test_moveBy_invalid_one_string_value(self):
        guideline = self.getGuideline_transform()
        with self.assertRaises(TypeError):
            guideline.moveBy((-1, "2"))

    def test_moveBy_invalid_all_strings_value(self):
        guideline = self.getGuideline_transform()
        with self.assertRaises(TypeError):
            guideline.moveBy("-1, 2")

    def test_moveBy_invalid_int_value(self):
        guideline = self.getGuideline_transform()
        with self.assertRaises(TypeError):
            guideline.moveBy(1)

    # scaleBy

    def test_scaleBy_valid_one_value_no_origin(self):
        guideline = self.getGuideline_transform()
        guideline.scaleBy(-2)
        self.assertEqual(guideline.x, -2)
        self.assertEqual(guideline.y, -4)
        self.assertAlmostEqual(guideline.angle, 225.000, places=3)

    def test_scaleBy_valid_two_values_no_origin(self):
        guideline = self.getGuideline_transform()
        guideline.scaleBy((-2, 3))
        self.assertEqual(guideline.x, -2)
        self.assertEqual(guideline.y, 6)
        self.assertAlmostEqual(guideline.angle, 123.690, places=3)

    def test_scaleBy_valid_two_values_origin(self):
        guideline = self.getGuideline_transform()
        guideline.scaleBy((-2, 3), origin=(1, 2))
        self.assertEqual(guideline.x, 1)
        self.assertEqual(guideline.y, 2)
        self.assertAlmostEqual(guideline.angle, 123.690, places=3)

    def test_scaleBy_invalid_one_string_value(self):
        guideline = self.getGuideline_transform()
        with self.assertRaises(TypeError):
            guideline.scaleBy((-1, "2"))

    def test_scaleBy_invalid_two_string_values(self):
        guideline = self.getGuideline_transform()
        with self.assertRaises(TypeError):
            guideline.scaleBy("-1, 2")

    def test_scaleBy_invalid_tuple_too_many_values(self):
        guideline = self.getGuideline_transform()
        with self.assertRaises(ValueError):
            guideline.scaleBy((-1, 2, -3))

    # rotateBy

    def test_rotateBy_valid_no_origin(self):
        guideline = self.getGuideline_transform()
        guideline.rotateBy(45)
        self.assertAlmostEqual(guideline.x, -0.707, places=3)
        self.assertAlmostEqual(guideline.y, 2.121, places=3)
        self.assertAlmostEqual(guideline.angle, 0.000, places=3)

    def test_rotateBy_valid_origin(self):
        guideline = self.getGuideline_transform()
        guideline.rotateBy(45, origin=(1, 2))
        self.assertAlmostEqual(guideline.x, 1)
        self.assertAlmostEqual(guideline.y, 2)
        self.assertAlmostEqual(guideline.angle, 0.000, places=3)

    def test_rotateBy_invalid_string_value(self):
        guideline = self.getGuideline_transform()
        with self.assertRaises(TypeError):
            guideline.rotateBy("45")

    def test_rotateBy_invalid_too_large_value_positive(self):
        guideline = self.getGuideline_transform()
        with self.assertRaises(ValueError):
            guideline.rotateBy(361)

    def test_rotateBy_invalid_too_large_value_negative(self):
        guideline = self.getGuideline_transform()
        with self.assertRaises(ValueError):
            guideline.rotateBy(-361)

    # skewBy

    def test_skewBy_valid_no_origin_one_value(self):
        guideline = self.getGuideline_transform()
        guideline.skewBy(100)
        self.assertAlmostEqual(guideline.x, -10.343, places=3)
        self.assertEqual(guideline.y, 2.0)
        self.assertAlmostEqual(guideline.angle, 8.525, places=3)

    def test_skewBy_valid_no_origin_two_values(self):
        guideline = self.getGuideline_transform()
        guideline.skewBy((100, 200))
        self.assertAlmostEqual(guideline.x, -10.343, places=3)
        self.assertAlmostEqual(guideline.y, 2.364, places=3)
        self.assertAlmostEqual(guideline.angle, 5.446, places=3)

    def test_skewBy_valid_origin_one_value(self):
        guideline = self.getGuideline_transform()
        guideline.skewBy(100, origin=(1, 2))
        self.assertEqual(guideline.x, 1)
        self.assertEqual(guideline.y, 2)
        self.assertAlmostEqual(guideline.angle, 8.525, places=3)

    def test_skewBy_valid_origin_two_values(self):
        guideline = self.getGuideline_transform()
        guideline.skewBy((100, 200), origin=(1, 2))
        self.assertEqual(guideline.x, 1)
        self.assertEqual(guideline.y, 2)
        self.assertAlmostEqual(guideline.angle, 5.446, places=3)

    # -------------
    # Normalization
    # -------------

    # round

    def getGuideline_round(self):
        guideline = self.getGuideline_generic()
        guideline.x = 1.1
        guideline.y = 2.5
        guideline.angle = 45.5
        return guideline

    def test_round_close_to(self):
        guideline = self.getGuideline_round()
        guideline.round()
        self.assertEqual(guideline.x, 1)

    def test_round_at_half(self):
        guideline = self.getGuideline_round()
        guideline.round()
        self.assertEqual(guideline.y, 3)

    def test_round_angle(self):
        guideline = self.getGuideline_round()
        guideline.round()
        self.assertEqual(guideline.angle, 45.5)

    # ----
    # Hash
    # ----

    def test_hash_object_self(self):
        guideline_one = self.getGuideline_generic()
        self.assertEqual(hash(guideline_one), hash(guideline_one))

    def test_hash_object_other(self):
        guideline_one = self.getGuideline_generic()
        guideline_two = self.getGuideline_generic()
        self.assertNotEqual(hash(guideline_one), hash(guideline_two))

    def test_hash_object_self_variable_assignment(self):
        guideline_one = self.getGuideline_generic()
        a = guideline_one
        self.assertEqual(hash(guideline_one), hash(a))

    def test_hash_object_other_variable_assignment(self):
        guideline_one = self.getGuideline_generic()
        guideline_two = self.getGuideline_generic()
        a = guideline_one
        self.assertNotEqual(hash(guideline_two), hash(a))

    def test_is_hashable(self):
        guideline_one = self.getGuideline_generic()
        self.assertTrue(isinstance(guideline_one, collections.abc.Hashable))

    # -------
    # Parents
    # -------

    def test_get_parent_font(self):
        font, _ = self.objectGenerator("font")
        layer = font.newLayer("L")
        glyph = layer.newGlyph("X")
        guideline = glyph.appendGuideline((0, 0), 90, "Test Guideline")
        self.assertIsNotNone(guideline.font)
        self.assertEqual(guideline.font, font)

    def test_get_parent_noFont(self):
        layer, _ = self.objectGenerator("layer")
        glyph = layer.newGlyph("X")
        guideline = glyph.appendGuideline((0, 0), 90, "Test Guideline")
        self.assertIsNone(guideline.font)

    def test_get_parent_layer(self):
        layer, _ = self.objectGenerator("layer")
        glyph = layer.newGlyph("X")
        guideline = glyph.appendGuideline((0, 0), 90, "Test Guideline")
        self.assertIsNotNone(guideline.layer)
        self.assertEqual(guideline.layer, layer)

    def test_get_parent_noLayer(self):
        glyph, _ = self.objectGenerator("glyph")
        guideline = glyph.appendGuideline((0, 0), 90, "Test Guideline")
        self.assertIsNone(guideline.font)
        self.assertIsNone(guideline.layer)

    def test_get_parent_glyph(self):
        glyph, _ = self.objectGenerator("glyph")
        guideline = glyph.appendGuideline((0, 0), 90, "Test Guideline")
        self.assertIsNotNone(guideline.glyph)
        self.assertEqual(guideline.glyph, glyph)

    def test_get_parent_noGlyph(self):
        guideline, _ = self.objectGenerator("guideline")
        self.assertIsNone(guideline.font)
        self.assertIsNone(guideline.layer)
        self.assertIsNone(guideline.glyph)

    def test_set_parent_glyph(self):
        glyph, _ = self.objectGenerator("glyph")
        guideline = self.getGuideline_generic()
        guideline.glyph = glyph
        self.assertIsNotNone(guideline.glyph)
        self.assertEqual(guideline.glyph, glyph)

    def test_set_parent_glyph_none(self):
        guideline, _ = self.objectGenerator("guideline")
        guideline.glyph = None
        self.assertIsNone(guideline.glyph)

    def test_set_parent_font_none(self):
        guideline, _ = self.objectGenerator("guideline")
        guideline.font = None
        self.assertIsNone(guideline.glyph)

    def test_set_parent_glyph_exists(self):
        glyph, _ = self.objectGenerator("glyph")
        otherGlyph, _ = self.objectGenerator("glyph")
        guideline = glyph.appendGuideline((0, 0), 90, "Test Guideline")
        with self.assertRaises(AssertionError):
            guideline.glyph = otherGlyph

    def test_set_parent_glyph_font_exists(self):
        guideline = self.getGuideline_fontGuideline()
        glyph, _ = self.objectGenerator("glyph")
        with self.assertRaises(AssertionError):
            guideline.glyph = glyph

    def test_set_parent_font_font_exists(self):
        guideline = self.getGuideline_fontGuideline()
        font, _ = self.objectGenerator("font")
        with self.assertRaises(AssertionError):
            guideline.font = font

    def test_set_parent_font_glyph_exists(self):
        guideline = self.getGuideline_glyphGuideline()
        font, _ = self.objectGenerator("font")
        with self.assertRaises(AssertionError):
            guideline.font = font

    # --------
    # Equality
    # --------

    def test_object_equal_self(self):
        guideline_one = self.getGuideline_generic()
        self.assertEqual(guideline_one, guideline_one)

    def test_object_not_equal_other(self):
        guideline_one = self.getGuideline_generic()
        guideline_two = self.getGuideline_generic()
        self.assertNotEqual(guideline_one, guideline_two)

    def test_object_equal_self_variable_assignment(self):
        guideline_one = self.getGuideline_generic()
        a = guideline_one
        a.x = 200
        self.assertEqual(guideline_one, a)

    def test_object_not_equal_other_variable_assignment(self):
        guideline_one = self.getGuideline_generic()
        guideline_two = self.getGuideline_generic()
        a = guideline_one
        self.assertNotEqual(guideline_two, a)

    # ---------
    # Selection
    # ---------

    def test_selected_true(self):
        guideline = self.getGuideline_generic()
        try:
            guideline.selected = False
        except NotImplementedError:
            return
        guideline.selected = True
        self.assertEqual(guideline.selected, True)

    def test_not_selected_false(self):
        guideline = self.getGuideline_generic()
        try:
            guideline.selected = False
        except NotImplementedError:
            return
        self.assertEqual(guideline.selected, False)



# pylint: disable=C0103, C0114
from __future__ import annotations

import datetime
from typing import Protocol, TypeVar

from fontTools.pens.basePen import AbstractPen
from fontTools.pens.pointPen import AbstractPointPen

# Generic
T = TypeVar("T")

CollectionType = list[T] | tuple[T, ...]

# Builtins
IntFloatType = int | float

# Point / Offset / Anchor (bPoints)
Coordinate = tuple[IntFloatType, IntFloatType]
CoordinateLike = list[IntFloatType] | Coordinate

# Bounding box — (xMin, yMin, xMax, yMax).
BoundingBox = tuple[float, float, float, float]
BoundingBoxLike = list[IntFloatType] | BoundingBox

# RGBA color — (r, g, b, a) in [0, 1]
RGBA = tuple[float, float, float, float]
RGBALike = list[IntFloatType] | RGBA

# Affine transformation matrix — (xx, xy, yx, yy, dx, dy).
AffineTransformation = tuple[float, float, float, float, float, float]
AffineTransformationLike = list[IntFloatType] | AffineTransformation

# Kerning pair — (first, second) glyph or group names.
KerningPair = tuple[str, str]
KerningPairLike = list[str] | KerningPair

# Scale factor — (sx, sy) multiplicative pair.
ScaleFactor = tuple[float, float]
ScaleFactorPair = list[IntFloatType] | ScaleFactor
ScaleFactorLike = IntFloatType | ScaleFactorPair

# Skew angle — (xAngle, yAngle) in degrees.
SkewAngle = tuple[float, float]
SkewAnglePair = list[IntFloatType] | SkewAngle
SkewAngleLike = IntFloatType | SkewAnglePair

# Interpolation factor — (xFactor, yFactor).
InterpolationFactor = tuple[float, float]
InterpolationFactorPair = list[IntFloatType] | InterpolationFactor
InterpolationFactorLike = IntFloatType | InterpolationFactorPair

# Compatibility
DiffType = list[tuple[int, str | None, str | None]]

# Pens
PenType = AbstractPen
PointPenType = AbstractPointPen

# Mapping
CharacterMappingType = dict[int, tuple[str, ...]]
ReverseComponentMappingType = dict[str, tuple[str, ...]]

# Kerning
KerningDictType = dict[KerningPair, KerningPair]

# Lib
LibValueType = (
    str
    | IntFloatType
    | bool
    | CollectionType["LibValueType"]
    | dict[str, "LibValueType"]
    | bytes
    | bytearray
    | datetime.datetime
)


class LibValue:
    # Documentation class for LibValueType
    """A :class:`~fontParts.base.BaseLib` value may be one of the following
    *non-collection* types:

    - :class:`str`
    - :class:`int`
    - :class:`float`
    - :class:`bool`
    - :class:`bytes`
    - :class:`bytearray`
    - :class:`datetime.datetime`

    In addition, a value may also be a :class:`list` or :class:`tuple` containing any of
    the types above, or a :class:`dict` mapping :class:`str` keys to values of those
    same types (including nested lists, tuples, or dicts).

    """


# Interpolation
InterpolatableType = TypeVar("InterpolatableType", bound="Interpolatable")


class Interpolatable(Protocol):
    """Represent a protocol for interpolatable types."""

    def __add__(
        self: InterpolatableType, other: InterpolatableType
    ) -> InterpolatableType: ...

    def __sub__(
        self: InterpolatableType, other: InterpolatableType
    ) -> InterpolatableType: ...

    def __mul__(
        self: InterpolatableType, other: InterpolationFactorLike
    ) -> InterpolatableType: ...



from __future__ import annotations
from typing import Any

import defcon
from fontParts.base import BaseInfo
from fontParts.fontshell.base import RBaseObject


class RInfo(RBaseObject, BaseInfo):
    wrapClass = defcon.Info

    def _getAttr(self, attr: str) -> Any:
        return getattr(self.naked(), attr)

    def _setAttr(self, attr: str, value: Any) -> None:
        setattr(self.naked(), attr, value)



from __future__ import annotations
from typing import TYPE_CHECKING, Any

import defcon
from fontParts.base import BaseLib
from fontParts.fontshell.base import RBaseObject

if TYPE_CHECKING:
    from collections.abc import ItemsView


class RLib(RBaseObject, BaseLib):
    wrapClass = defcon.Lib

    def _items(self) -> ItemsView[str, Any]:
        return self.naked().items()

    def _contains(self, key: str) -> bool:
        return key in self.naked()

    def _setItem(self, key: str, value: Any) -> None:
        self.naked()[key] = value

    def _getItem(self, key: str) -> Any:
        return self.naked()[key]

    def _delItem(self, key: str) -> None:
        del self.naked()[key]



import sys
import unittest
from fontParts.test import test_normalizers
from fontParts.test import test_font
from fontParts.test import test_info
from fontParts.test import test_groups
from fontParts.test import test_kerning
from fontParts.test import test_features
from fontParts.test import test_layer
from fontParts.test import test_glyph
from fontParts.test import test_contour
from fontParts.test import test_segment
from fontParts.test import test_bPoint
from fontParts.test import test_point
from fontParts.test import test_component
from fontParts.test import test_anchor
from fontParts.test import test_image
from fontParts.test import test_lib
from fontParts.test import test_guideline
from fontParts.test import test_deprecated
from fontParts.test import test_color
from fontParts.test import test_bounds
from fontParts.test import test_world


def testEnvironment(objectGenerator, inApp=False, verbosity=1, testNormalizers=True):
    modules = [
        test_font,
        test_info,
        test_groups,
        test_kerning,
        test_features,
        test_layer,
        test_glyph,
        test_contour,
        test_segment,
        test_bPoint,
        test_point,
        test_component,
        test_anchor,
        test_image,
        test_lib,
        test_guideline,
        test_deprecated,
        test_color,
        test_bounds,
        test_world,
    ]
    if testNormalizers:
        modules.append(test_normalizers)

    globalSuite = unittest.TestSuite()
    loader = unittest.TestLoader()
    for module in modules:
        suite = loader.loadTestsFromModule(module)
        _setObjectGenerator(suite, objectGenerator)
        globalSuite.addTest(suite)
    runner = unittest.TextTestRunner(verbosity=verbosity)
    succes = runner.run(globalSuite).wasSuccessful()
    if not inApp:
        sys.exit(not succes)
    else:
        return succes  # pragma: no cover


def _setObjectGenerator(suite, objectGenerator):
    for i in suite:
        if isinstance(i, unittest.TestSuite):
            _setObjectGenerator(i, objectGenerator)
        else:
            i.objectGenerator = objectGenerator



import unittest
import tempfile
import os
from fontParts.world import RFont, FontList, OpenFont, OpenFonts


class TestFontList(unittest.TestCase):
    def getFont(self):
        font, _ = self.objectGenerator("font")
        return font

    # ----
    # Sort
    # ----

    def getFonts_sortBy(self, attr, values):
        fonts = FontList()
        for value in values:
            font = self.getFont()
            setattr(font.info, attr, value)
            if attr != "familyName":
                font.info.familyName = f"{attr} {repr(value)}"
            fonts.append(font)
        return fonts

    # Sort Description Strings
    # ------------------------

    def getFont_sortBy_monospaceGlyphs(self):
        font = self.getFont()
        font.info.familyName = f"monospace {str(id(font))}"
        glyph1 = font.newGlyph("a")
        glyph1.width = 100
        glyph2 = font.newGlyph("b")
        glyph2.width = 100
        return font

    def getFont_sortBy_proportionalGlyphs(self):
        font = self.getFont()
        font.info.familyName = f"proportional {str(id(font))}"
        glyph1 = font.newGlyph("a")
        glyph1.width = 100
        glyph2 = font.newGlyph("b")
        glyph2.width = 200
        return font

    # familyName

    def test_sortBy_familyName(self):
        fonts = self.getFonts_sortBy("familyName", ["aaa", "bbb", "ccc", None])
        font1, font2, font3, font4 = fonts
        fonts.sortBy("familyName")
        expected = [font4, font1, font2, font3]
        self.assertEqual(fonts, expected)

    # styleName

    def test_sortBy_styleName(self):
        fonts = self.getFonts_sortBy("styleName", ["aaa", "bbb", "ccc", None])
        font1, font2, font3, font4 = fonts
        fonts.sortBy("styleName")
        expected = [font4, font1, font2, font3]
        self.assertEqual(fonts, expected)

    # isRoman

    def test_sortBy_isRoman_styleMapStyleName(self):
        fonts = self.getFonts_sortBy(
            "styleMapStyleName", ["regular", "italic", "bold", "bold italic"]
        )
        font1, font2, font3, font4 = fonts
        fonts.reverse()
        fonts.sortBy("isRoman")
        expected = [font3, font1, font4, font2]
        self.assertEqual(fonts, expected)

    def test_sortBy_isRoman_italicAngle(self):
        fonts = self.getFonts_sortBy("italicAngle", [1, 2, 3, 0])
        font1, font2, font3, font4 = fonts
        fonts.sortBy("isRoman")
        expected = [font4, font1, font2, font3]
        self.assertEqual(fonts, expected)

    # isItalic

    def test_sortBy_isItalic_styleMapStyleName(self):
        fonts = self.getFonts_sortBy(
            "styleMapStyleName", ["regular", "italic", "bold", "bold italic"]
        )
        font1, font2, font3, font4 = fonts
        fonts.sortBy("isItalic")
        expected = [font2, font4, font1, font3]
        self.assertEqual(fonts, expected)

    def test_sortBy_isItalic_italicAngle(self):
        fonts = self.getFonts_sortBy("italicAngle", [0, 1, 2, 3])
        font1, font2, font3, font4 = fonts
        fonts.sortBy("isItalic")
        expected = [font2, font3, font4, font1]
        self.assertEqual(fonts, expected)

    # widthValue

    def test_sortBy_widthValue(self):
        fonts = self.getFonts_sortBy("openTypeOS2WidthClass", [1, 2, 3, None])
        font1, font2, font3, font4 = fonts
        fonts.sortBy("widthValue")
        expected = [font4, font1, font2, font3]
        self.assertEqual(fonts, expected)

    # weightValue

    def test_sortBy_weightValue(self):
        fonts = self.getFonts_sortBy("openTypeOS2WeightClass", [100, 200, 300, None])
        font1, font2, font3, font4 = fonts
        fonts.sortBy("weightValue")
        expected = [font4, font1, font2, font3]
        self.assertEqual(fonts, expected)

    # isMonospace

    def test_sortBy_isMonospace_postscriptIsFixedPitch(self):
        fonts = self.getFonts_sortBy(
            "postscriptIsFixedPitch", [True, True, False, False]
        )
        font1, font2, font3, font4 = fonts
        fonts.reverse()
        fonts.sortBy("isMonospace")
        expected = [font2, font1, font4, font3]
        self.assertEqual(fonts, expected)

    def test_sortBy_isMonospace_glyphs(self):
        font1 = self.getFont_sortBy_monospaceGlyphs()
        font2 = self.getFont_sortBy_monospaceGlyphs()
        font3 = self.getFont_sortBy_proportionalGlyphs()
        font4 = self.getFont_sortBy_proportionalGlyphs()
        fonts = FontList()
        fonts.extend([font1, font2, font3, font4])
        fonts.reverse()
        fonts.sortBy("isMonospace")
        expected = [font2, font1, font4, font3]
        self.assertEqual(fonts, expected)

    # isProportional

    def test_sortBy_isProportional_postscriptIsFixedPitch(self):
        fonts = self.getFonts_sortBy(
            "postscriptIsFixedPitch", [False, False, True, True]
        )
        font1, font2, font3, font4 = fonts
        fonts.reverse()
        fonts.sortBy("isProportional")
        expected = [font2, font1, font4, font3]
        self.assertEqual(fonts, expected)

    def test_sortBy_isProportional_glyphs(self):
        font1 = self.getFont_sortBy_monospaceGlyphs()
        font2 = self.getFont_sortBy_monospaceGlyphs()
        font3 = self.getFont_sortBy_proportionalGlyphs()
        font4 = self.getFont_sortBy_proportionalGlyphs()
        fonts = FontList()
        fonts.extend([font1, font2, font3, font4])
        fonts.sortBy("isProportional")
        expected = [font3, font4, font1, font2]
        self.assertEqual(fonts, expected)

    # font.info Attributes

    def test_sortBy_fontInfoAttribute_xHeight(self):
        fonts = self.getFonts_sortBy("xHeight", [10, 20, 30, 40])
        font1, font2, font3, font4 = fonts
        fonts.reverse()
        fonts.sortBy("xHeight")
        expected = [font1, font2, font3, font4]
        self.assertEqual(fonts, expected)

    # Sort Value Function
    # -------------------

    def getFont_withGlyphCount(self, count):
        font = self.getFont()
        for i in range(count):
            font.newGlyph(f"glyph{i}")
        font.info.familyName = str(count)
        return font

    def test_sortBy_sortValueFunction(self):
        font1 = self.getFont_withGlyphCount(10)
        font2 = self.getFont_withGlyphCount(20)
        font3 = self.getFont_withGlyphCount(30)
        font4 = self.getFont_withGlyphCount(40)
        fonts = FontList()
        fonts.extend([font1, font2, font3, font4])
        fonts.reverse()

        def glyphCountSortValue(font):
            return len(font)

        fonts.sortBy(glyphCountSortValue)
        expected = [font1, font2, font3, font4]
        self.assertEqual(fonts, expected)

    # ------
    # Search
    # ------

    # family name

    def test_getFontsByFamilyName(self):
        font1 = self.getFont()
        font1.info.familyName = "A"
        font2 = self.getFont()
        font2.info.familyName = "B"
        font3 = self.getFont()
        font3.info.familyName = "C"
        font4 = self.getFont()
        font4.info.familyName = "A"
        fonts = FontList()
        fonts.extend([font1, font2, font3, font4])
        found = fonts.getFontsByFamilyName("A")
        self.assertEqual(found, [font1, font4])

    # style name

    def test_getFontsByStyleName(self):
        font1 = self.getFont()
        font1.info.styleName = "A"
        font2 = self.getFont()
        font2.info.styleName = "B"
        font3 = self.getFont()
        font3.info.styleName = "C"
        font4 = self.getFont()
        font4.info.styleName = "A"
        fonts = FontList()
        fonts.extend([font1, font2, font3, font4])
        found = fonts.getFontsByStyleName("A")
        self.assertEqual(found, [font1, font4])

    # family name, style name

    def test_getFontsByFamilyNameStyleName(self):
        font1 = self.getFont()
        font1.info.familyName = "A"
        font1.info.styleName = "1"
        font2 = self.getFont()
        font2.info.familyName = "A"
        font2.info.styleName = "2"
        font3 = self.getFont()
        font3.info.familyName = "B"
        font3.info.styleName = "1"
        font4 = self.getFont()
        font4.info.familyName = "A"
        font4.info.styleName = "1"
        fonts = FontList()
        fonts.extend([font1, font2, font3, font4])
        found = fonts.getFontsByFamilyNameStyleName("A", "1")
        self.assertEqual(found, [font1, font4])


class TestFontOpen(unittest.TestCase):
    def setUp(self):
        font, _ = self.objectGenerator("font")
        self.font_dir = tempfile.mkdtemp()
        self.font_path = os.path.join(self.font_dir, "test.ufo")
        font.save(self.font_path)

    def tearDown(self):
        import shutil

        shutil.rmtree(self.font_dir)

    def test_font_open(self):
        OpenFont(self.font_path)


class TestOpenFonts(unittest.TestCase):
    def setUp(self):
        self.font_dir = tempfile.mkdtemp()
        for i in range(3):
            font, _ = self.objectGenerator("font")
            path = os.path.join(self.font_dir, f"test{i}.ufo")
            font.save(path)

    def tearDown(self):
        import shutil

        shutil.rmtree(self.font_dir)

    def test_font_open(self):
        fonts = OpenFonts(self.font_dir)
        fileNames = [os.path.basename(font.path) for font in fonts]
        fileNames.sort()
        expected = ["test0.ufo", "test1.ufo", "test2.ufo"]
        self.assertEqual(fileNames, expected)


class TestFontShell_RFont(unittest.TestCase):
    def test_fontshell_RFont_empty(self):
        RFont()



import unittest
from fontParts.base.color import Color


class TestColor(unittest.TestCase):
    def test_new_tuple(self):
        color = Color((0.1, 0.2, 0.3, 1.0))
        self.assertEqual(color.r, 0.1)
        self.assertEqual(color.g, 0.2)
        self.assertEqual(color.b, 0.3)
        self.assertEqual(color.a, 1.0)

    def test_new_args(self):
        color = Color((0.1, 0.2, 0.3, 1.0))
        self.assertEqual(color.r, 0.1)
        self.assertEqual(color.g, 0.2)
        self.assertEqual(color.b, 0.3)
        self.assertEqual(color.a, 1.0)

    def test_new_kwargs(self):
        color = Color(r=0.1, g=0.2, b=0.3, a=1.0)
        self.assertEqual(color.r, 0.1)
        self.assertEqual(color.g, 0.2)
        self.assertEqual(color.b, 0.3)
        self.assertEqual(color.a, 1.0)



from __future__ import annotations

import defcon

from fontParts.base.annotations import IntFloatType
from fontParts.base import BasePoint, FontPartsError
from fontParts.fontshell.base import RBaseObject


class RPoint(RBaseObject, BasePoint):
    wrapClass = defcon.Point

    def _init(self, pathOrObject: defcon.Point | None = None) -> None:
        if pathOrObject is None and self.wrapClass is not None:
            pathOrObject = self.wrapClass((0, 0))
        super()._init(pathOrObject=pathOrObject)

    def _postChangeNotification(self) -> None:
        contour = self.contour
        if contour is None:
            return
        contour.naked().postNotification("Contour.PointsChanged")
        self.changed()

    def changed(self) -> None:
        self.contour.naked().dirty = True

    # ----------
    # Attributes
    # ----------

    # type

    def _get_type(self) -> str:
        value = self.naked().segmentType
        if value is None:
            value = "offcurve"
        return value

    def _set_type(self, value: str) -> None:
        self.naked().segmentType = None if value == "offcurve" else value
        self._postChangeNotification()

    # smooth

    def _get_smooth(self) -> bool:
        return self.naked().smooth

    def _set_smooth(self, value: bool) -> None:
        self.naked().smooth = value
        self._postChangeNotification()

    # x

    def _get_x(self) -> IntFloatType:
        return self.naked().x

    def _set_x(self, value: IntFloatType) -> None:
        self.naked().x = value
        self._postChangeNotification()

    # y

    def _get_y(self) -> IntFloatType:
        return self.naked().y

    def _set_y(self, value: IntFloatType) -> None:
        self.naked().y = value
        self._postChangeNotification()

    # --------------
    # Identification
    # --------------

    # name

    def _get_name(self) -> str | None:
        return self.naked().name

    def _set_name(self, value: str) -> None:
        self.naked().name = value
        self._postChangeNotification()

    # identifier

    def _get_identifier(self) -> str | None:
        return self.naked().identifier

    def _getIdentifier(self) -> str:
        point = self.naked()
        value = point.identifier
        if value is not None:
            return value
        if self.contour is not None:
            contour = self.contour.naked()
            contour.generateIdentifierForPoint(point)
            value = point.identifier
        else:
            raise FontPartsError(
                "An identifier can not be generated "
                "for this point because it does not "
                "belong to a contour."
            )
        return value



from __future__ import annotations

from collections import namedtuple
from fontParts.base.annotations import IntFloatType
from typing import overload, Any

from fontParts.base.annotations import BoundingBoxLike
from fontParts.base.normalizers import normalizeBoundingBox

_BaseBounds = namedtuple("_BaseBounds", ["xMin", "yMin", "xMax", "yMax"])


class Bounds(_BaseBounds):
    """Represent a bounding box object.

    This class accepts either four coordinate values or a :class:`list` or
    :class:`tuple` of these values.

    :param xMin: The bound's minimum x value as :class:`int` or :class:`float`.
    :param yMin: The bound's maximum y value as :class:`int` or :class:`float`.
    :param xMax: The bound's maximum x value as :class:`int` or :class:`float`.
    :param yMax: The bound's minimum y value as :class:`int` or :class:`float`.

    """

    __slots__ = ()

    @overload
    def __new__(
        cls,
        xMin: IntFloatType,
        yMin: IntFloatType,
        xMax: IntFloatType,
        yMax: IntFloatType,
    ) -> Bounds: ...

    @overload
    def __new__(cls, bounds: BoundingBoxLike) -> Bounds: ...

    def __new__(cls, *args: Any, **kwargs: Any) -> Bounds:
        if kwargs:
            value = (kwargs["xMin"], kwargs["yMin"], kwargs["xMax"], kwargs["yMax"])
        else:
            value = args[0] if len(args) == 1 else args
        normalizedValue = normalizeBoundingBox(value)
        return super().__new__(cls, *normalizedValue)

    @property
    def width(self) -> IntFloatType:
        """Get the width of the bounds.

        :return: A :class:`float` representing the width of the bounds.

        """
        return self.xMax - self.xMin

    @property
    def height(self) -> float:
        """Get the height of the bounds.

        :return: A :class:`float` representing the height of the bounds.

        """
        return self.yMax - self.yMin

    @property
    def xCenter(self) -> float:
        """Get the center of the x-axis.

        :return: A :class:`float` representing the center of the x-axis

        """
        return self.xMin + (self.width / 2)

    @property
    def yCenter(self) -> float:
        """Get the center of the y-axis.

        :return: A :class:`float` representing the center of the y-axis

        """
        return self.yMin + (self.height / 2)

    # Remove auto-generated _BaseCpøpr.__new__ docstring
    __new__.__doc__ = ""



from __future__ import annotations
from typing import TYPE_CHECKING, Any

import defcon
import booleanOperations
from fontParts.base import BaseGlyph
from fontParts.base.annotations import (
    AffineTransformationLike,
    RGBALike,
    RGBA,
    BoundingBox,
    CoordinateLike,
    CollectionType,
    IntFloatType,
)
from fontParts.base.errors import FontPartsError
from fontParts.fontshell.base import RBaseObject
from fontParts.fontshell.contour import RContour
from fontParts.fontshell.component import RComponent
from fontParts.fontshell.anchor import RAnchor
from fontParts.fontshell.guideline import RGuideline
from fontParts.fontshell.image import RImage
from fontParts.fontshell.lib import RLib
from fontTools.ufoLib.glifLib import (
    GlifLibError,
    readGlyphFromString,
    writeGlyphToString,
)

if TYPE_CHECKING:
    from fontTools.pens.pointPen import SegmentToPointPen
    from defcon.pens.glyphObjectPointPen import (
        GlyphObjectPointPen,
        GlyphObjectLoadingPointPen,
    )


class RGlyph(RBaseObject, BaseGlyph):
    wrapClass = defcon.Glyph
    contourClass = RContour
    componentClass = RComponent
    anchorClass = RAnchor
    guidelineClass = RGuideline
    imageClass = RImage
    libClass = RLib

    # --------------
    # Identification
    # --------------

    # Name

    def _get_name(self) -> str:
        return self.naked().name

    def _set_name(self, value: str) -> None:
        self.naked().name = value

    # Unicodes

    def _get_unicodes(self) -> CollectionType[int]:
        return self.naked().unicodes

    def _set_unicodes(self, value: CollectionType[int]) -> None:
        self.naked().unicodes = value

    # -------
    # Metrics
    # -------

    # horizontal

    def _get_width(self) -> IntFloatType:
        return self.naked().width

    def _set_width(self, value: IntFloatType) -> None:
        self.naked().width = value

    def _get_leftMargin(self) -> IntFloatType | None:
        return self.naked().leftMargin

    def _set_leftMargin(self, value: IntFloatType) -> None:
        self.naked().leftMargin = value

    def _get_rightMargin(self) -> IntFloatType | None:
        return self.naked().rightMargin

    def _set_rightMargin(self, value: IntFloatType) -> None:
        self.naked().rightMargin = value

    # vertical

    def _get_height(self) -> IntFloatType:
        return self.naked().height

    def _set_height(self, value: IntFloatType) -> None:
        self.naked().height = value

    def _get_bottomMargin(self) -> IntFloatType | None:
        return self.naked().bottomMargin

    def _set_bottomMargin(self, value: IntFloatType) -> None:
        self.naked().bottomMargin = value

    def _get_topMargin(self) -> IntFloatType | None:
        return self.naked().topMargin

    def _set_topMargin(self, value: IntFloatType) -> None:
        self.naked().topMargin = value

    # ------
    # Bounds
    # ------

    def _get_bounds(self) -> BoundingBox | None:
        return self.naked().bounds

    # ----
    # Area
    # ----

    def _get_area(self) -> float | None:
        value = self.naked().area
        if not value:
            return None
        return value

    # ----
    # Pens
    # ----

    def getPen(self) -> SegmentToPointPen:
        return self.naked().getPen()

    def getPointPen(self) -> GlyphObjectPointPen | GlyphObjectLoadingPointPen:
        return self.naked().getPointPen()

    # -----------------------------------------
    # Contour, Component and Anchor Interaction
    # -----------------------------------------

    # Contours

    def _lenContours(self, **kwargs: Any) -> int:
        return len(self.naked())

    def _getContour(self, index: int, **kwargs: Any) -> RContour:
        contour = self.naked()[index]
        return self.contourClass(contour)

    def _removeContour(self, index: int, **kwargs: Any) -> None:
        glyph = self.naked()
        contour = glyph[index]
        glyph.removeContour(contour)

    def _removeOverlap(self, **kwargs: Any) -> None:
        if len(self):
            contours = list(self)
            for contour in contours:
                for point in contour.points:
                    if point.type == "qcurve":
                        raise TypeError("fontshell can't removeOverlap for quadratics")
            self.clear(
                contours=True,
                components=False,
                anchors=False,
                guidelines=False,
                image=False,
            )
            booleanOperations.union(contours, self.getPointPen())

    def _correctDirection(self, trueType: bool = False, **kwargs: Any) -> None:
        self.naked().correctContourDirection(trueType=trueType)

    # Components

    def _lenComponents(self, **kwargs: Any) -> int:
        return len(self.naked().components)

    def _getComponent(self, index: int, **kwargs: Any) -> RComponent:
        component = self.naked().components[index]
        return self.componentClass(component)

    def _removeComponent(self, index: int, **kwargs: Any) -> None:
        glyph = self.naked()
        component = glyph.components[index]
        glyph.removeComponent(component)

    # Anchors

    def _lenAnchors(self, **kwargs: Any) -> int:
        return len(self.naked().anchors)

    def _getAnchor(self, index: int, **kwargs: Any) -> RAnchor:
        anchor = self.naked().anchors[index]
        return self.anchorClass(anchor)

    def _appendAnchor(
        self,
        name: str,
        position: CoordinateLike | None = None,
        color: RGBALike | None = None,
        identifier: str | None = None,
        **kwargs: Any,
    ) -> RAnchor:
        glyph = self.naked()
        anchor = self.anchorClass().naked()
        if anchor is None:
            raise ValueError("Anchor cannot be None")
        anchor.name = name
        if position is not None:
            anchor.x = position[0]
            anchor.y = position[1]
        anchor.color = color
        anchor.identifier = identifier
        glyph.appendAnchor(anchor)
        wrapped = self.anchorClass(anchor)
        wrapped.glyph = self
        return wrapped

    def _removeAnchor(self, index: int, **kwargs: Any) -> None:
        glyph = self.naked()
        anchor = glyph.anchors[index]
        glyph.removeAnchor(anchor)

    # Guidelines

    def _lenGuidelines(self, **kwargs: Any) -> int:
        return len(self.naked().guidelines)

    def _getGuideline(self, index: int, **kwargs: Any) -> RGuideline:
        guideline = self.naked().guidelines[index]
        return self.guidelineClass(guideline)

    def _appendGuideline(
        self,
        position: CoordinateLike | None,
        angle: float,
        name: str | None = None,
        color: RGBALike | None = None,
        identifier: str | None = None,
        **kwargs: Any,
    ) -> RGuideline:
        glyph = self.naked()
        guideline = self.guidelineClass().naked()
        if guideline is None:
            raise ValueError("Guideline cannot be None")
        if position is not None:
            guideline.x = position[0]
            guideline.y = position[1]
        guideline.angle = angle
        guideline.name = name
        guideline.color = color
        guideline.identifier = identifier
        glyph.appendGuideline(guideline)
        return self.guidelineClass(guideline)

    def _removeGuideline(self, index: int, **kwargs: Any) -> None:
        glyph = self.naked()
        guideline = glyph.guidelines[index]
        glyph.removeGuideline(guideline)

    # -----------------
    # Layer Interaction
    # -----------------

    # new

    def _newLayer(self, name: str, **kwargs: Any) -> RGlyph:
        layerName = name
        glyphName = self.name
        font = self.font
        if layerName not in font.layerOrder:
            layer = font.newLayer(layerName)
        else:
            layer = font.getLayer(layerName)
        glyph = layer.newGlyph(glyphName)
        return glyph

    # remove

    def _removeLayer(self, name: str, **kwargs: Any) -> None:
        layerName = name
        glyphName = self.name
        font = self.font
        layer = font.getLayer(layerName)
        layer.removeGlyph(glyphName)

    # -----
    # Image
    # -----

    def _get_image(self) -> RImage | None:
        image = self.naked().image
        if image is None:
            return None
        return self.imageClass(image)

    def _addImage(
        self,
        data: bytes,
        transformation: AffineTransformationLike | None = None,
        color: RGBALike | None = None,
    ) -> RImage:
        image = self.naked().image
        image = self.imageClass(image)
        image.glyph = self
        image.data = data
        image.transformation = transformation
        image.color = color
        return image

    def _clearImage(self, **kwargs: Any) -> None:
        self.naked().image = None

    # ----
    # Note
    # ----

    # Mark

    def _get_markColor(self) -> RGBA | None:
        value = self.naked().markColor
        if value is not None:
            value = tuple(value)
        return value

    def _set_markColor(self, value: RGBALike | None) -> None:
        self.naked().markColor = value

    # Note

    def _get_note(self) -> str | None:
        return self.naked().note

    def _set_note(self, value: str | None) -> None:
        self.naked().note = value

    # -----------
    # Sub-Objects
    # -----------

    # lib

    def _get_lib(self) -> RLib:
        return self.libClass(pathOrObject=self.naked().lib)

    # tempLib

    def _get_tempLib(self) -> RLib:
        return self.libClass(pathOrObject=self.naked().tempLib)

    # ---
    # API
    # ---

    def _loadFromGLIF(self, glifData: str, validate: bool = True) -> None:
        try:
            readGlyphFromString(
                aString=glifData,
                glyphObject=self.naked(),
                pointPen=self.getPointPen(),
                validate=validate,
            )
        except GlifLibError as e:
            raise FontPartsError("Not valid glif data") from e

    def _dumpToGLIF(self, glyphFormatVersion: int) -> str:
        glyph = self.naked()
        return writeGlyphToString(
            glyphName=glyph.name,
            glyphObject=glyph,
            drawPointsFunc=glyph.drawPoints,
            formatVersion=glyphFormatVersion,
        )



from __future__ import annotations
from typing import TYPE_CHECKING, Any
from collections.abc import Callable

from fontTools.misc import transform
from fontParts.base.base import (
    BaseObject,
    TransformationMixin,
    SelectionMixin,
    dynamicProperty,
    reference,
)
from fontParts.base import normalizers
from fontParts.base.color import Color
from fontParts.base.deprecated import DeprecatedImage, RemovedImage
from fontParts.base.annotations import (
    ScaleFactorLike,
    ScaleFactorPair,
    ScaleFactor,
    AffineTransformationLike,
    AffineTransformation,
    Coordinate,
    CoordinateLike,
    RGBA,
    RGBALike,
)

if TYPE_CHECKING:
    from fontParts.base.glyph import BaseGlyph
    from fontParts.base.layer import BaseLayer
    from fontParts.base.font import BaseFont


class BaseImage(
    BaseObject, TransformationMixin, SelectionMixin, DeprecatedImage, RemovedImage
):
    """Represent the basis for an image object."""

    copyAttributes = ("transformation", "color", "data")

    def _reprContents(self):
        contents = [f"offset='({self.offset[0]}, {self.offset[1]})'"]
        if self.color:
            contents.append(f"color={self.color!r}")
        if self.glyph is not None:
            contents.append("in glyph")
            contents += self.glyph._reprContents()
        return contents

    def __bool__(self) -> bool:
        if self.data is None:
            return False
        elif len(self.data) == 0:
            return False
        else:
            return True

    __nonzero__ = __bool__

    # -------
    # Parents
    # -------

    # Glyph

    _glyph: Callable[[], BaseGlyph] | None = None

    glyph = dynamicProperty(
        "glyph",
        """Get or set the image's parent glyph object.

        The value must be a :class:`BaseGlyph` instance or :obj:`None`.

        :return: The :class:`BaseGlyph` instance containing the image
            or :obj:`None`.
        :raises AssertionError: If attempting to set the glyph when it
            has already been set.

        Example::

            >>> glyph = image.glyph

        """,
    )

    def _get_glyph(self) -> BaseGlyph | None:
        if self._glyph is None:
            return None
        return self._glyph()

    def _set_glyph(self, glyph: BaseGlyph | Callable[[], BaseGlyph] | None) -> None:
        if self._glyph is not None:
            raise AssertionError("glyph for image already set")
        if glyph is not None:
            glyph = reference(glyph)
        self._glyph = glyph

    # Layer

    layer: dynamicProperty = dynamicProperty(
        "layer",
        """Get the image's parent layer object.

        This property is read-only.

        :return: The :class:`BaseLayer` instance containing the image
            or :obj:`None`.

        Example::

            >>> layer = image.layer

        """,
    )

    def _get_layer(self) -> BaseLayer | None:
        if self._glyph is None:
            return None
        return self.glyph.layer

    # Font

    font: dynamicProperty = dynamicProperty(
        "font",
        """Get the image's parent font object.

        This property is read-only.

        :return: The :class:`BaseFont` instance containing the image
            or :obj:`None`.

        Example::

            >>> font = image.font

        """,
    )

    def _get_font(self) -> BaseFont | None:
        if self._glyph is None:
            return None
        return self.glyph.font

    # ----------
    # Attributes
    # ----------

    # Transformation

    transformation: dynamicProperty = dynamicProperty(
        "base_transformation",
        """Get or set the image's transformation matrix.

        The value must be a :ref:`type-transformation`.

        :return: A :ref:`type-transformation` value representing the
            transformation matrix of the image.

        Example::

            >>> image.transformation
            (1, 0, 0, 1, 0, 0)
            >>> image.transformation = (2, 0, 0, 2, 100, -50)
            
        """,
    )

    def _get_base_transformation(self) -> AffineTransformation:
        value = self._get_transformation()
        value = normalizers.normalizeTransformationMatrix(value)
        return value

    def _set_base_transformation(self, value: AffineTransformationLike) -> None:
        value = normalizers.normalizeTransformationMatrix(value)
        self._set_transformation(value)

    def _get_transformation(self) -> AffineTransformation:
        """Get the native image's transformation matrix.

        This is the environment implementation of the
        :attr:`BaseImage.transformation` property getter.

        :return: A :ref:`type-transformation` value representing the
            transformation matrix of the image. The value will be
            normalized with :func:`normalizers.normalizeTransformationMatrix`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_transformation(self, value: AffineTransformationLike) -> None:
        """Set the native image's transformation matrix.

        This is the environment implementation of the
        :attr:`BaseImage.transformation` property setter.

        :param value: The :ref:`type-transformation` to set. The value will have
            been normalized with :func:`normalizers.normalizeTransformationMatrix`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    offset: dynamicProperty = dynamicProperty(
        "base_offset",
        """Get or set the image's offset.

        The value must be a :ref:`type-coordinate`.

        :return: A :ref:`type-coordinate` representing the offset of the image.

        Example::

            >>> image.offset
            (0, 0)
            >>> image.offset = (100, -50)

        """,
    )

    def _get_base_offset(self) -> Coordinate:
        value = self._get_offset()
        value = normalizers.normalizeTransformationOffset(value)
        return value

    def _set_base_offset(self, value: CoordinateLike) -> None:
        value = normalizers.normalizeTransformationOffset(value)
        self._set_offset(value)

    def _get_offset(self) -> Coordinate:
        """Get the native image's offset.

        This is the environment implementation of the :attr:`BaseImage.offset`
        property getter.

        :return: A :ref:`type-coordinate` representing the offset of the image.
            The value will be normalized
            with :func:`normalizers.normalizeTransformationOffset`.

        .. note::

            Subclasses may override this method.

        """
        sx, sxy, syx, sy, ox, oy = self.transformation
        return (ox, oy)

    def _set_offset(self, value: CoordinateLike) -> None:
        """Set the native image's offset.

        This is the environment implementation of the :attr:`BaseImage.offset`
        property setter.

        :param value: The offset to set as a :ref:`type-coordinate`. The value will
            have been normalized with :func:`normalizers.normalizeTransformationOffset`.

        .. note::

            Subclasses may override this method.

        """
        sx, sxy, syx, sy, ox, oy = self.transformation
        ox, oy = value
        self.transformation = (sx, sxy, syx, sy, ox, oy)

    scale: dynamicProperty = dynamicProperty(
        "base_scale",
        """Get or set the image's scale.

        The value must be a :class:`list` or :class:`tuple` of two :class:`int`
        or :class:`float` items representing the ``(x, y)`` scale of the image.

        :return: A :class:`tuple` of two :class:`float` items representing the
            ``(x, y)`` scale of the image.

        Example::

            >>> image.scale
            (1, 1)
            >>> image.scale = (2, 2)
            
        """,
    )

    def _get_base_scale(self) -> ScaleFactor:
        value = self._get_scale()
        value = normalizers.normalizeTransformationScale(value)
        return value

    def _set_base_scale(self, value: ScaleFactorLike) -> None:
        value = normalizers.normalizeTransformationScale(value)
        self._set_scale(value)

    def _get_scale(self) -> ScaleFactor:
        """Get the native image's scale.

        This is the environment implementation of the :attr:`BaseImage.scale`
        property getter.

        :return: A :class:`tuple` of two :class:`float` items representing the
            ``(x, y)`` scale of the image. The value will have been normalized
            with :func:`normalizers.normalizeTransformationScale`.

        .. note::

            Subclasses may override this method.

        """
        sx, sxy, syx, sy, ox, oy = self.transformation
        return (sx, sy)

    def _set_scale(self, value: ScaleFactorPair) -> None:
        """Set the native image's scale.

        This is the environment implementation of the :attr:`BaseImage.scale`
        property setter.

        :param value: The scale to set as a :class:`list` or :class:`tuple`
            of :class:`int` or :class:`float` items representing the ``(x, y)``
            scale of the image. The value will have been normalized
            with :func:`normalizers.normalizeTransformationScale`.

        .. note::

            Subclasses may override this method.

        """
        sx, sxy, syx, sy, ox, oy = self.transformation
        sx, sy = value
        self.transformation = (sx, sxy, syx, sy, ox, oy)

    # Color

    color: dynamicProperty = dynamicProperty(
        "base_color",
        """Get or set the image's color.

        The value must be a :ref:`type-color` or :obj:`None`.

        :return: A :class:`Color` instance representing the color of the image,
            or :obj:`None`.

        Example::

            >>> image.color
            None
            >>> image.color = (1, 0, 0, 0.5)

        """,
    )

    def _get_base_color(self) -> Color | None:
        value = self._get_color()
        if value is not None:
            value = Color(value)
        return value

    def _set_base_color(self, value: RGBALike | None) -> None:
        if value is not None:
            value = normalizers.normalizeColor(value)
        self._set_color(value)

    def _get_color(self) -> RGBA | None:
        """Get the native image's color.

        This is the environment implementation of the :attr:`BaseImage.color`
        property getter.

        :return: A :ref:`type-color` representing the color of the image,
            or :obj:`None`. The value will be normalized
            with :func:`normalizers.normalizeColor`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_color(self, value: RGBALike | None) -> None:
        """Set the native image's color.

        This is the environment implementation of the :attr:`BaseImage.color`
        property setter.

        :param value: The :ref:`type-color` to set for the image or :obj:`None`.
            The value will have been normalized with :func:`normalizers.normalizeColor`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # Data

    data: dynamicProperty = dynamicProperty(
        "base_data",
        """Get or set the image's raw byte data.
        
        The possible formats are defined by each environment.
        The value must be a :class:`bytes` object.

        :return: A :class:`bytes` object representing the raw byte data of the image 
            or :obj:`None`.

        """,
    )

    def _get_base_data(self) -> bytes | None:
        return self._get_data()

    def _set_base_data(self, value: bytes) -> None:
        self._set_data(value)

    def _get_data(self) -> bytes | None:
        """Get the native image's raw byte data.

        This is the environment implementation of the :attr:`BaseImage.data`
        property getter.

        :return: A :class:`bytes` object representing the raw byte data of the image
            or :obj:`None`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_data(self, value: bytes) -> None:
        """Set the native image's raw byte data.

        This is the environment implementation of the :attr:`BaseImage.data`
        property setter.

        :param value: The :class:`bytes` object to set for the image.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # --------------
    # Transformation
    # --------------

    def _transformBy(self, matrix: AffineTransformationLike, **kwargs: Any) -> None:
        r"""Transform the native image.

        This is the environment implementation of :meth:`BaseImage.transformBy`.

        :param matrix: The transformation to apply as a :ref:`type-transformation`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        t = transform.Transform(*matrix)
        transformation = t.transform(self.transformation)
        self.transformation = tuple(transformation)

    # -------------
    # Normalization
    # -------------

    def round(self) -> None:
        """Round the image's offset coordinates.

        Example::

            >>> image.round()

        """
        self._round()

    def _round(self) -> None:
        """Round the native image's offset coordinates.

        This is the environment implementation of :meth:`BaseImage.round`.

        .. note::

            Subclasses may override this method.

        """
        x, y = self.offset
        x = normalizers.normalizeVisualRounding(x)
        y = normalizers.normalizeVisualRounding(y)
        self.offset = (x, y)



from __future__ import annotations

import defcon
from fontParts.base import BaseAnchor
from fontParts.base.annotations import RGBALike, RGBA
from fontParts.fontshell.base import RBaseObject


class RAnchor(RBaseObject, BaseAnchor):
    wrapClass = defcon.Anchor

    def _init(self, pathOrObject: defcon.Anchor | None = None) -> None:
        if self.wrapClass is not None:
            if pathOrObject is None:
                pathOrObject = self.wrapClass()
                pathOrObject.x = 0
                pathOrObject.y = 0
            super()._init(pathOrObject=pathOrObject)

    # --------
    # Position
    # --------

    # x

    def _get_x(self) -> float:
        return self.naked().x

    def _set_x(self, value: float) -> None:
        self.naked().x = value

    # y

    def _get_y(self) -> float:
        return self.naked().y

    def _set_y(self, value: float) -> None:
        self.naked().y = value

    # --------------
    # Identification
    # --------------

    # identifier

    def _get_identifier(self) -> str | None:
        return self.naked().identifier

    def _getIdentifier(self) -> str:
        return self.naked().generateIdentifier()

    def _setIdentifier(self, value: str) -> None:
        self.naked().identifier = value

    # name

    def _get_name(self) -> str | None:
        return self.naked().name

    def _set_name(self, value: str | None) -> None:
        self.naked().name = value

    # color

    def _get_color(self) -> RGBA | None:
        value = self.naked().color
        if value is not None:
            value = tuple(value)
        return value

    def _set_color(self, value: RGBALike | None) -> None:
        self.naked().color = value



from fontParts.base.errors import FontPartsError
from fontParts.base.font import BaseFont
from fontParts.base.info import BaseInfo
from fontParts.base.groups import BaseGroups
from fontParts.base.kerning import BaseKerning
from fontParts.base.features import BaseFeatures
from fontParts.base.lib import BaseLib
from fontParts.base.layer import BaseLayer
from fontParts.base.glyph import BaseGlyph
from fontParts.base.contour import BaseContour
from fontParts.base.point import BasePoint
from fontParts.base.segment import BaseSegment
from fontParts.base.bPoint import BaseBPoint
from fontParts.base.component import BaseComponent
from fontParts.base.anchor import BaseAnchor
from fontParts.base.guideline import BaseGuideline
from fontParts.base.image import BaseImage



from __future__ import annotations
from typing import TYPE_CHECKING, Any
from collections.abc import Callable
from fontTools.misc import transform

from fontParts.base.base import (
    BaseObject,
    TransformationMixin,
    SelectionMixin,
    IdentifierMixin,
    dynamicProperty,
    reference,
)
from fontParts.base.errors import FontPartsError
from fontParts.base import normalizers
from fontParts.base.deprecated import DeprecatedBPoint, RemovedBPoint
from fontParts.base.annotations import (
    AffineTransformationLike,
    Coordinate,
    CoordinateLike,
)

if TYPE_CHECKING:
    from fontParts.base.contour import BaseContour
    from fontParts.base.font import BaseFont
    from fontParts.base.glyph import BaseGlyph
    from fontParts.base.layer import BaseLayer
    from fontParts.base.point import BasePoint
    from fontParts.base.segment import BaseSegment


class BaseBPoint(
    BaseObject,
    TransformationMixin,
    SelectionMixin,
    DeprecatedBPoint,
    IdentifierMixin,
    RemovedBPoint,
):
    """Represent the basis for a bPoint object."""

    def _reprContents(self) -> list[str]:
        contents = [f"{self.type}", f"anchor='({self.anchor[0]}, {self.anchor[1]})'"]
        return contents

    def _setPoint(self, point: BasePoint) -> None:
        if hasattr(self, "_point"):
            raise AssertionError("point for bPoint already set")
        self._point = point

    def __eq__(self, other):
        if hasattr(other, "_point"):
            return self._point == other._point
        return NotImplemented

    # this class should not be used in hashable
    # collections since it is dynamically generated.

    __hash__ = None  # type: ignore[assignment]

    # -------
    # Parents
    # -------

    # identifier

    def _get_identifier(self) -> str | None:
        """Get the native bPoint's unique identifier.

        This is the environment implementation of :attr:`BaseBPoint.identifier`.

        If the native object does not have an identifier assigned, one may be
        assigned with :meth:`BaseBPoint.getIdentifier`

        :return: The unique identifier assigned to the object as a :class:`str`,
            or :obj:`None` indicating the object has no identifier.

        .. note::

            Subclasses may override this method.

        """
        return self._point.identifier

    def _getIdentifier(self) -> str:
        """Generate and assign a unique identifier to the native bPoint.

        This is the environment implementation of :meth:`BaseBPoint.getIdentifier`.

        :return: A unique object identifier as a :class:`str`.

        .. note::

            Subclasses may override this method.

        """
        return self._point.getIdentifier()

    # Segment

    _segment: dynamicProperty = dynamicProperty("base_segment")

    def _get_base_segment(self) -> BaseSegment | None:
        point = self._point
        for segment in self.contour.segments:
            if segment.onCurve == point:
                return segment
        return None

    _nextSegment: dynamicProperty = dynamicProperty("base_nextSegment")

    def _get_base_nextSegment(self) -> BaseSegment | None:
        contour = self.contour
        if contour is None:
            return None
        segments = contour.segments
        segment = self._segment
        i = segments.index(segment) + 1
        if i >= len(segments):
            i = i % len(segments)
        nextSegment = segments[i]
        return nextSegment

    # Contour

    _contour: Callable[[], BaseContour] | None = None

    contour = dynamicProperty(
        "contour",
        """Get or set the bPoint's parent contour object.

        The value must be a :class:`BaseContour` instance or :obj:`None`.

        :return: The :class:`BaseContour` instance containing the bPoint
            or :obj:`None`.
        :raises AssertionError: If attempting to set the contour when it
            has already been set.

        Example::

            >>> contour = bPoint.contour

        """,
    )

    def _get_contour(self) -> BaseContour | None:
        if self._contour is None:
            return None
        return self._contour()

    def _set_contour(
        self, contour: BaseContour | Callable[[], BaseContour] | None
    ) -> None:
        if self._contour is not None:
            raise AssertionError("contour for bPoint already set")
        if contour is not None:
            contour = reference(contour)
        self._contour = contour

    # Glyph

    glyph: dynamicProperty = dynamicProperty(
        "glyph",
        """Get the bPoint's parent glyph object.

        This property is read-only.

        The value must be a :class:`BaseGlyph` instance or :obj:`None`.

        :return: The :class:`BaseGlyph` instance containing the bPoint
            or :obj:`None`.

        Example::

            >>> glyph = bPoint.glyph

        """,
    )

    def _get_glyph(self) -> BaseGlyph | None:
        if self._contour is None:
            return None
        return self.contour.glyph

    # Layer

    layer: dynamicProperty = dynamicProperty(
        "layer",
        """Get the bPoint's parent layer object.

        This property is read-only.

        :return: The :class:`BaseLayer` instance containing the bPoint
            or :obj:`None`.

        Example::

            >>> layer = bPoint.layer

        """,
    )

    def _get_layer(self) -> BaseLayer | None:
        if self._contour is None:
            return None
        return self.glyph.layer

    # Font

    font: dynamicProperty = dynamicProperty(
        "font",
        """Get the bPoint's parent font object.

        This property is read-only.

        :return: The :class:`BaseFont` instance containing the bPoint
            or :obj:`None`.

        Example::

            >>> font = bPoint.font

        """,
    )

    def _get_font(self) -> BaseFont | None:
        if self._contour is None:
            return None
        return self.glyph.font

    # ----------
    # Attributes
    # ----------

    # anchor

    anchor: dynamicProperty = dynamicProperty(
        "base_anchor",
        """Get or set the bPoint's anchor point.

        The value must be a :ref:`type-coordinate`.

        :return: a :ref:`type-coordinate` representing the anchor point of the bPoint.

        """,
    )

    def _get_base_anchor(self) -> Coordinate:
        value = self._get_anchor()
        value = normalizers.normalizeCoordinateTuple(value)
        return value

    def _set_base_anchor(self, value: CoordinateLike) -> None:
        value = normalizers.normalizeCoordinateTuple(value)
        self._set_anchor(value)

    def _get_anchor(self) -> Coordinate:
        """Get the bPoint's anchor point.

        This is the environment implementation of the :attr:`BaseBPoint.anchor`
        property getter.

        :return: a :ref:`type-coordinate` representing the anchor point of the
            bPoint. The value will be normalized
            with :func:`normalizers.normalizeCoordinateTuple`.

        .. note::

            Subclasses may override this method.

        """
        point = self._point
        return (point.x, point.y)

    def _set_anchor(self, value: CoordinateLike) -> None:
        """Set the bPoint's anchor point.

        This is the environment implementation of the :attr:`BaseBPoint.anchor`
        property setter.

        :param value: The anchor point to set as a :ref:`type-coordinate`.
            The value will have been normalized
            with :func:`normalizers.normalizeCoordinateTuple`.

        .. note::

            Subclasses may override this method.

        """
        pX, pY = self.anchor
        x, y = value
        dX = x - pX
        dY = y - pY
        self.moveBy((dX, dY))

    # bcp in

    bcpIn: dynamicProperty = dynamicProperty(
        "base_bcpIn",
        """Get or set the bPoint's incoming off-curve.

        The value must be a :ref:`type-coordinate`.

        :return: A :ref:`type-coordinate` representing the incoming
            off-curve of the bPoin.

        """,
    )

    def _get_base_bcpIn(self) -> Coordinate:
        value = self._get_bcpIn()
        value = normalizers.normalizeCoordinateTuple(value)
        return value

    def _set_base_bcpIn(self, value: CoordinateLike) -> None:
        value = normalizers.normalizeCoordinateTuple(value)
        self._set_bcpIn(value)

    def _get_bcpIn(self) -> Coordinate:
        """Get the bPoint's incoming off-curve.

        This is the environment implementation of the :attr:`BaseBPoint.bcpIn`
        property getter.

        :return: A :ref:`type-coordinate` representing the incoming off-curve of
            the bPoin. The value will be normalized
            with :func:`normalizers.normalizeCoordinateTuple`.

        .. note::

            Subclasses may override this method.

        """
        segment = self._segment
        offCurves = segment.offCurve
        if offCurves:
            bcp = offCurves[-1]
            x, y = relativeBCPIn(self.anchor, (bcp.x, bcp.y))
        else:
            x = y = 0
        return (x, y)

    def _set_bcpIn(self, value: CoordinateLike) -> None:
        """Set the bPoint's incoming off-curve.

        This is the environment implementation of the :attr:`BaseBPoint.bcpIn`
        property setter.

        :param value: The incoming off-curve to set as
            a :ref:`type-coordinate`. The value will have been normalized
            with :func:`normalizers.normalizeCoordinateTuple`.
        :raises FontPartsError: When attempting to set the incoming off-curve
            for a the first point in an open contour.

        .. note::

            Subclasses may override this method.

        """
        x, y = absoluteBCPIn(self.anchor, value)
        segment = self._segment
        if segment.type == "move" and value != (0, 0):
            raise FontPartsError(
                "Cannot set the bcpIn for the first point in an open contour."
            )

        offCurves = segment.offCurve
        if offCurves:
            # if the two off-curves are located at the anchor
            # coordinates we can switch to a line segment type.
            if value == (0, 0) and self.bcpOut == (0, 0):
                segment.type = "line"
                segment.smooth = False
            else:
                offCurves[-1].x = x
                offCurves[-1].y = y
        elif value != (0, 0):
            segment.type = "curve"
            offCurves = segment.offCurve
            offCurves[-1].x = x
            offCurves[-1].y = y

    # bcp out

    bcpOut: dynamicProperty = dynamicProperty(
        "base_bcpOut",
        """Get or set the bPoint's outgoing off-curve.

        The value must be a :ref:`type-coordinate`.

        :return: A :ref:`type-coordinate` representing the outgoing
            off-curve of the bPoin.

        """,
    )

    def _get_base_bcpOut(self) -> Coordinate:
        value = self._get_bcpOut()
        value = normalizers.normalizeCoordinateTuple(value)
        return value

    def _set_base_bcpOut(self, value: CoordinateLike) -> None:
        value = normalizers.normalizeCoordinateTuple(value)
        self._set_bcpOut(value)

    def _get_bcpOut(self) -> Coordinate:
        """Get the bPoint's outgoing off-curve.

        This is the environment implementation of the :attr:`BaseBPoint.bcpOut`
        property getter.

        :return: A :ref:`type-coordinate` representing the outgoing
            off-curve of the bPoin. The value will be normalized
            with :func:`normalizers.normalizeCoordinateTuple`.

        .. note::

            Subclasses may override this method.

        """
        nextSegment = self._nextSegment
        offCurves = nextSegment.offCurve
        if offCurves:
            bcp = offCurves[0]
            x, y = relativeBCPOut(self.anchor, (bcp.x, bcp.y))
        else:
            x = y = 0
        return (x, y)

    def _set_bcpOut(self, value: CoordinateLike) -> None:
        """Set the bPoint's outgoing off-curve.

        This is the environment implementation of the :attr:`BaseBPoint.bcpOut`
        property setter.

        :param value: The outgoing off-curve to set as
            a :ref:`type-coordinate`. The value will have been normalized
            with :func:`normalizers.normalizeCoordinateTuple`.
        :raises FontPartsError: When attempting to set the outgoing off-curve
            for a the last point in an open contour.

        .. note::

            Subclasses may override this method.

        """
        x, y = absoluteBCPOut(self.anchor, value)
        segment = self._segment
        nextSegment = self._nextSegment
        if nextSegment.type == "move" and value != (0, 0):
            raise FontPartsError(
                "Cannot set the bcpOut for the last point in an open contour."
            )
        else:
            offCurves = nextSegment.offCurve
            if offCurves:
                # if the off-curves are located at the anchor coordinates
                # we can switch to a "line" segment type
                if value == (0, 0) and self.bcpIn == (0, 0):
                    segment.type = "line"
                    segment.smooth = False
                else:
                    offCurves[0].x = x
                    offCurves[0].y = y
            elif value != (0, 0):
                nextSegment.type = "curve"
                offCurves = nextSegment.offCurve
                offCurves[0].x = x
                offCurves[0].y = y

    # type

    type: dynamicProperty = dynamicProperty(
        "base_type",
        """Get or set the bPoint's type.

        The value must be a :class:`str` containing one of the following
        alternatives:

        +--------------+-----------------------------------------------------------+
        | Type         | Description                                               |
        +--------------+-----------------------------------------------------------+
        | ``'curve'``  | A point where bcpIn and bcpOut are smooth (linked).       |
        | ``'corner'`` | A point where bcpIn and bcpOut are not smooth (unlinked). |
        +--------+-----------------------------------------------------------------+

        :return: A :class:`str` representing the type of the bPoint.

        """,
    )

    def _get_base_type(self) -> str:
        value = self._get_type()
        value = normalizers.normalizeBPointType(value)
        return value

    def _set_base_type(self, value: str) -> None:
        value = normalizers.normalizeBPointType(value)
        self._set_type(value)

    def _get_type(self) -> str:
        """Get the bPoint's type.

        This is the environment implementation of the :attr:`BaseBPoint.type`
        property getter.

        :return: A :class:`str` representing the type of the bPoint. The value
            will be normalized with :func:`normalizers.normalizeBPointType`.
        :raises FontPartsError: If the point's type cannot be converted to a valid
            bPoint type.

        .. note::

            Subclasses may override this method.

        """
        point = self._point
        typ = point.type
        bType = None
        if point.smooth:
            if typ == "curve":
                bType = "curve"
            elif typ == "line" or typ == "move":
                nextSegment = self._nextSegment
                if nextSegment is not None and nextSegment.type == "curve":
                    bType = "curve"
                else:
                    bType = "corner"
        elif typ in ("move", "line", "curve"):
            bType = "corner"

        if bType is None:
            raise FontPartsError(f"A {typ} point can not be converted to a bPoint.")
        return bType

    def _set_type(self, value: str) -> None:
        """Set the bPoint's type.

        This is the environment implementation of the :attr:`BaseBPoint.type`
        property setter.

        :param value: The outgoing off-curve to set as a :class:`str`. The value
            will have been normalized with :func:`normalizers.normalizeBPointType`.

        .. note::

            Subclasses may override this method.

        """
        point = self._point
        # convert corner to curve
        if value == "curve" and point.type == "line":
            # This needs to insert off-curves without
            # generating unnecessary points in the
            # following segment. The segment object
            # implements this logic, so delegate the
            # change to the corresponding segment.
            segment = self._segment
            segment.type = "curve"
            segment.smooth = True
        # convert curve to corner
        elif value == "corner" and point.type == "curve":
            point.smooth = False

    # --------------
    # Identification
    # --------------

    index: dynamicProperty = dynamicProperty(
        "base_index",
        """Get the index of the bPoint.

        This property is read-only.

        :return: An :class:`int` representing the bPoint's index within an
            ordered list of the parent contour's bPoints, or :obj:`None` if the
            bPoint does not belong to a contour.

        Example::

            >>> bPoint.index
            0

        """,
    )

    def _get_base_index(self) -> int | None:
        if self.contour is None:
            return None
        value = self._get_index()
        value = normalizers.normalizeIndex(value)
        return value

    def _get_index(self) -> int | None:
        """Get the index of the native bPoint.

        This is the environment implementation of the :attr:`BaseBPoint.index`
        property getter.

        :return: An :class:`int` representing the bPoint's index within an
            ordered list of the parent contour's bPoints, or :obj:`None` if the
            bPoint does not belong to a contour. The value will be
            normalized with :func:`normalizers.normalizeIndex`.

        .. note::

            Subclasses may override this method.

        """
        contour = self.contour
        value = contour.bPoints.index(self)
        return value

    # --------------
    # Transformation
    # --------------

    def _transformBy(self, matrix: AffineTransformationLike, **kwargs: Any) -> None:
        r"""Transform the native bPoint.

        This is the environment implementation of :meth:`BaseBPoint.transformBy`.

        :param matrix: The transformation to apply as a :ref:`type-transformation`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        anchor = self.anchor
        bcpIn = absoluteBCPIn(anchor, self.bcpIn)
        bcpOut = absoluteBCPOut(anchor, self.bcpOut)
        points = [bcpIn, anchor, bcpOut]
        t = transform.Transform(*matrix)
        bcpIn, anchor, bcpOut = t.transformPoints(points)
        x, y = anchor
        self._point.x = x
        self._point.y = y
        self.bcpIn = relativeBCPIn(anchor, bcpIn)
        self.bcpOut = relativeBCPOut(anchor, bcpOut)

    # ----
    # Misc
    # ----

    def round(self) -> None:
        """Round the bPoint's coordinates.

        This applies to:

        - :attr:`anchor`
        - :attr:`bcpIn`
        - :attr:`bcpOut`

        Example::

            >>> bPoint.round()

        """
        x, y = self.anchor
        self.anchor = (
            normalizers.normalizeVisualRounding(x),
            normalizers.normalizeVisualRounding(y),
        )
        x, y = self.bcpIn
        self.bcpIn = (
            normalizers.normalizeVisualRounding(x),
            normalizers.normalizeVisualRounding(y),
        )
        x, y = self.bcpOut
        self.bcpOut = (
            normalizers.normalizeVisualRounding(x),
            normalizers.normalizeVisualRounding(y),
        )


def relativeBCPIn(anchor: CoordinateLike, BCPIn: CoordinateLike) -> Coordinate:
    """convert absolute incoming bcp value to a relative value.

    :param anchor: The anchor reference point from which to measure the relative
        BCP value as a :ref:`type-coordinate`.
    :param BCPIn: The absolute incoming BCP value to be converted as
        a :ref:`type-coordinate`.
    :return: The relative position of the incoming BCP as a :ref:`type-coordinate`.

    """
    return (BCPIn[0] - anchor[0], BCPIn[1] - anchor[1])


def absoluteBCPIn(anchor: CoordinateLike, BCPIn: CoordinateLike) -> Coordinate:
    """convert relative incoming bcp value to an absolute value.

    :param anchor: The anchor reference point from which the relative BCP value
        is measured as a :ref:`type-coordinate`.
    :param BCPIn: The relative incoming BCP value to be converted as
        a :ref:`type-coordinate`.
    :return: The absolute position of the incoming BCP as a :ref:`type-coordinate`.

    """
    return (BCPIn[0] + anchor[0], BCPIn[1] + anchor[1])


def relativeBCPOut(anchor: CoordinateLike, BCPOut: CoordinateLike) -> Coordinate:
    """convert absolute outgoing bcp value to a relative value.

    :param anchor: The anchor reference point from which to measure the relative
        BCP value as a :ref:`type-coordinate`.
    :param BCPOut: The absolute outgoing BCP value to be converted as
        a :ref:`type-coordinate`.
    :return: The relative position of the outgoing BCP as a :ref:`type-coordinate`.

    """
    return (BCPOut[0] - anchor[0], BCPOut[1] - anchor[1])


def absoluteBCPOut(anchor: CoordinateLike, BCPOut: CoordinateLike) -> Coordinate:
    """convert relative outgoing bcp value to an absolute value.

    :param anchor: The anchor reference point from which the relative BCP value
        is measured as a :ref:`type-coordinate`.
    :param BCPOut: The relative outgoing BCP value to be converted as
        a :ref:`type-coordinate`.
    :return: The absolute position of the outgoing BCP as a :ref:`type-coordinate`.

    """
    return (BCPOut[0] + anchor[0], BCPOut[1] + anchor[1])



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from fontTools.misc import transform

from fontParts.base import normalizers
from fontParts.base.annotations import AffineTransformationLike, IntFloatType
from fontParts.base.base import (
    BaseObject,
    IdentifierMixin,
    PointPositionMixin,
    SelectionMixin,
    TransformationMixin,
    dynamicProperty,
    reference,
)
from fontParts.base.deprecated import DeprecatedPoint, RemovedPoint

if TYPE_CHECKING:
    from fontParts.base.contour import BaseContour


class BasePoint(
    BaseObject,
    TransformationMixin,
    PointPositionMixin,
    SelectionMixin,
    IdentifierMixin,
    DeprecatedPoint,
    RemovedPoint,
):
    """Represent the basis for a point object.

    This object is almost always created with :meth:`BaseContour.appendPoint`,
    the pen returned by :meth:`BaseGlyph.getPen` or the point pen returned
    by :meth:`BaseGlyph.getPointPen`.

    An orphan point can be created like this::

        >>> point = RPoint()

    """

    copyAttributes: tuple[str, str, str, str, str] = (
        "type",
        "smooth",
        "x",
        "y",
        "name",
    )

    def _reprContents(self) -> list[str]:
        contents = [f"{self.type}", f"({self.x}, {self.y})"]
        if self.name is not None:
            contents.append(f"name='{self.name}'")
        if self.smooth:
            contents.append(f"smooth={self.smooth!r}")
        return contents

    # -------
    # Parents
    # -------

    # Contour

    _contour: Callable[[], BaseContour] | None = None

    contour: dynamicProperty = dynamicProperty(
        "contour",
        """Get or set the point's parent contour object.

        The value must be a :class:`BaseContour` instance or :obj:`None`.

        :return: The :class:`BaseContour` instance containing the point
            or :obj:`None`.
        :raises AssertionError: If attempting to set the contour when it
            has already been set.

        Example::

            >>> contour = point.contour

        """,
    )

    def _get_contour(self) -> BaseContour | None:
        if self._contour is None:
            return None
        return self._contour()

    def _set_contour(
        self, contour: BaseContour | Callable[[], BaseContour] | None
    ) -> None:
        if self._contour is not None:
            raise AssertionError("contour for point already set")
        if contour is not None:
            contour = reference(contour)
        self._contour = contour

    # Glyph

    glyph: dynamicProperty = dynamicProperty(
        "glyph",
        """Get the point's parent glyph object.

        This property is read-only.

        The value must be a :class:`BaseGlyph` instance or :obj:`None`.

        :return: The :class:`BaseGlyph` instance containing the point
            or :obj:`None`.

        Example::

            >>> glyph = point.glyph

        """,
    )

    def _get_glyph(self) -> BaseObject | None:
        if self._contour is None:
            return None
        return self.contour.glyph

    # Layer

    layer: dynamicProperty = dynamicProperty(
        "layer",
        """Get the point's parent layer object.

        This property is read-only.

        :return: The :class:`BaseLayer` instance containing the point
            or :obj:`None`.

        Example::

            >>> layer = point.layer

        """,
    )

    def _get_layer(self) -> BaseObject | None:
        if self._contour is None:
            return None
        return self.glyph.layer

    # Font

    font: dynamicProperty = dynamicProperty(
        "font",
        """Get the point's parent font object.

        This property is read-only.

        :return: The :class:`BaseFont` instance containing the point
            or :obj:`None`.

        Example::

            >>> font = point.font

        """,
    )

    def _get_font(self) -> BaseObject | None:
        if self._contour is None:
            return None
        return self.glyph.font

    # ----------
    # Attributes
    # ----------

    # type

    type: dynamicProperty = dynamicProperty(
        "base_type",
        """Get or set the point's type.

        The value must be a :class:`str` containing one of the following
        alternatives:

        +----------------+---------------------------------+
        | Type           | Description                     |
        +----------------+---------------------------------+
        | ``'move'``     | An on-curve move to.            |
        | ``'line'``     | An on-curve line to.            |
        | ``'curve'``    | An on-curve cubic curve to.     |
        | ``'qcurve'``   | An on-curve quadratic curve to. |
        | ``'offcurve'`` | An off-curve.                   |
        +----------------+---------------------------------+

        :return: A :class:`str` representing the type of the point.

        """,
    )

    def _get_base_type(self) -> str:
        value = self._get_type()
        value = normalizers.normalizePointType(value)
        return value

    def _set_base_type(self, value: str) -> None:
        value = normalizers.normalizePointType(value)
        self._set_type(value)

    def _get_type(self) -> str:  # type: ignore[return]
        """Get the native point's type.

        This is the environment implementation of the :attr:`BasePoint.type`
        property getter.

        :return: A :class:`str` representing the type of the point. The value
            will be normalized with :func:`normalizers.normalizePointType`.
        :raises NotImplementedError: If the method has not been overridden by
            a subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_type(self, value: str) -> None:
        """Set the native point's type.

        Description

        This is the environment implementation of the :attr:`BasePoint.type`
        property setter.

        :param value: The point type definition as a :class:`str`. The value
            will have been normalized with :func:`normalizers.normalizePointType`.
        :raises NotImplementedError: If the method has not been overridden by
            a subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # smooth

    smooth: dynamicProperty = dynamicProperty(
        "base_smooth",
        """Get or set the point's smooth state.

        The value must be a :class:`bool` indicating the point's smooth state.

        :return: :obj:`True` if the point is smooth, :obj:`False` if it is sharp.

        Example::

            >>> point.smooth
            False
            >>> point.smooth = True

        """,
    )

    def _get_base_smooth(self) -> bool:
        value = self._get_smooth()
        value = normalizers.normalizeBoolean(value)
        return value

    def _set_base_smooth(self, value: bool) -> None:
        value = normalizers.normalizeBoolean(value)
        self._set_smooth(value)

    def _get_smooth(self) -> bool:  # type: ignore[return]
        """Get the native point's smooth state.

        This is the environment implementation of the :attr:`BasePoint.smooth`
        property getter.

        :return: A :class:`bool` indicating the point's smooth state. The value
            will be normalized with :func:`normalizers.normalizeBoolean`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_smooth(self, value: bool) -> None:
        """Set the native point's smooth state.

        This is the environment implementation of the :attr:`BasePoint.smooth`
        property setter.

        :param value: The point's smooth state as a :class:`bool`. The value
            will have been normalized with :func:`normalizers.normalizeBoolean`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # x

    x: dynamicProperty = dynamicProperty(
        "base_x",
        """Get or set the x coordinate of the point.

        The value must be an :class:`int` or a :class:`float`.

        :return: An :class:`int` or a :class:`float` representing the point's
            x coordinate.

        Example::

            >>> point.x
            100
            >>> point.x = 101

        """,
    )

    def _get_base_x(self) -> IntFloatType:
        value = self._get_x()
        value = normalizers.normalizeX(value)
        return value

    def _set_base_x(self, value: IntFloatType) -> None:
        value = normalizers.normalizeX(value)
        self._set_x(value)

    def _get_x(self) -> IntFloatType:  # type: ignore[return]
        """Get the x coordinate of the native point.

        Description

        This is the environment implementation of the :attr:`BasePoint.x`
        property getter.

        :return: An :class:`int` or a :class:`float` representing the point's
            x coordinate. The value will be normalized with
            :func:`normalizers.normalizeX`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_x(self, value: IntFloatType) -> None:
        """Set the x coordinate of the native point.

        This is the environment implementation of the :attr:`BasePoint.x`
        property setter.

        :param value: The point's x coodinate to set as an :class:`int`
            or :class:`float`. The value will have been normalized
            with :func:`normalizers.normalizeX`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # y

    y: dynamicProperty = dynamicProperty(
        "base_y",
        """Get or set the y coordinate of the point.

        The value must be an :class:`int` or a :class:`float`.

        :return: An :class:`int` or a :class:`float` representing the point's
            y coordinate.

        Example::

            >>> point.y
            100
            >>> point.y = 101

        """,
    )

    def _get_base_y(self) -> IntFloatType:
        value = self._get_y()
        value = normalizers.normalizeY(value)
        return value

    def _set_base_y(self, value: IntFloatType) -> None:
        value = normalizers.normalizeY(value)
        self._set_y(value)

    def _get_y(self) -> IntFloatType:  # type: ignore[return]
        """Get the y coordinate of the native point.

        This is the environment implementation of the :attr:`BasePoint.y`
        property getter.

        :return: An :class:`int` or a :class:`float` representing the point's
            y coordinate. The value will be normalized
            with :func:`normalizers.normalizeY`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_y(self, value: IntFloatType) -> None:
        """Set the y coordinate of the native point.

        This is the environment implementation of the :attr:`BasePoint.y`
        property setter.

        :param value: The point's y coordinate as an :class:`int`
            or :class:`float`. The value will have been normalized
            with :func:`normalizers.normalizeY`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # --------------
    # Identification
    # --------------

    # index

    index: dynamicProperty = dynamicProperty(
        "base_index",
        """Get the index of the point.

        This property is read-only.

        :return: An :class:`int` representing the point's index within an
            ordered list of the parent contour's points, or :obj:`None` if the
            point does not belong to a contour.

        Example::

            >>> point.index
            0

        """,
    )

    def _get_base_index(self) -> int | None:
        value = self._get_index()
        value = normalizers.normalizeIndex(value)
        return value

    def _get_index(self) -> int | None:
        """Get the index of the native point.

        This is the environment implementation of the :attr:`BasePoint.index`
        property getter.

        :return: An :class:`int` representing the point's index within an
            ordered list of the parent contour's points, or :obj:`None` if the
            point does not belong to a contour. The value will be
            normalized with :func:`normalizers.normalizeIndex`.

        .. note::

            Subclasses may override this method.

        """
        contour = self.contour
        if contour is None:
            return None
        return contour.points.index(self)

    # name

    name: dynamicProperty = dynamicProperty(
        "base_name",
        """Get or set the name of the point.

        The value must be a :class:`str` or :obj:`None`.

        :return: A :class:`str` representing the point's name or :obj:`None`.

        Example::

            >>> point.name
            'my point'
            >>> point.name = None

        """,
    )

    def _get_base_name(self) -> str | None:
        value = self._get_name()
        if value is not None:
            value = normalizers.normalizePointName(value)
        return value

    def _set_base_name(self, value: str) -> None:
        if value is not None:
            value = normalizers.normalizePointName(value)
        self._set_name(value)

    def _get_name(self) -> str | None:  # type: ignore[return]
        """Get the name of the native point.

        This is the environment implementation of the :attr:`BasePoint.name`
        property getter.

        :return: A :class:`str` representing the point's name or :obj:`None`.
            The value will be normalized with :func:`normalizers.normalizePointName`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_name(self, value: str) -> None:
        """Set the name of the native point.

        This is the environment implementation of the :attr:`BasePoint.name`
        property setter.

        :param value: The point name as a :class:`str`. The value
            will have been normalized with :func:`normalizers.normalizePointName`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # --------------
    # Transformation
    # --------------

    def _transformBy(self, matrix: AffineTransformationLike, **kwargs: Any) -> None:
        r"""Transform the native point.

        This is the environment implementation of :meth:`BasePoint.transformBy`.

        :param matrix: The transformation to apply as a :ref:`type-transformation`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        t = transform.Transform(*matrix)
        x, y = t.transformPoint((self.x, self.y))
        self.x = x
        self.y = y

    # -------------
    # Normalization
    # -------------

    def round(self) -> None:
        """Round the point's coordinates.

        This applies to:

        - :attr:`x`
        - :attr:`y`

        Example::

            >>> point.round()

        """
        self._round()

    def _round(self, **kwargs: Any) -> None:
        r"""Round the native point's coordinates.

        This is the environment implementation of :meth:`BasePoint.round`.

        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        self.x = normalizers.normalizeVisualRounding(self.x)
        self.y = normalizers.normalizeVisualRounding(self.y)



from __future__ import annotations
from typing import Any
import os

import defcon
from fontParts.base.annotations import RGBALike, CoordinateLike, CollectionType
from fontParts.base import BaseFont
from fontParts.fontshell.base import RBaseObject
from fontParts.fontshell.info import RInfo
from fontParts.fontshell.groups import RGroups
from fontParts.fontshell.kerning import RKerning
from fontParts.fontshell.features import RFeatures
from fontParts.fontshell.lib import RLib
from fontParts.fontshell.layer import RLayer
from fontParts.fontshell.guideline import RGuideline


class RFont(RBaseObject, BaseFont):
    wrapClass = defcon.Font
    infoClass = RInfo
    groupsClass = RGroups
    kerningClass = RKerning
    featuresClass = RFeatures
    libClass = RLib
    layerClass = RLayer
    guidelineClass = RGuideline

    # ---------------
    # File Operations
    # ---------------

    # Initialize

    def _init(
        self,
        pathOrObject: str | os.PathLike | defcon.Font | None = None,
        showInterface: bool = True,
        **kwargs: Any,
    ) -> None:
        if self.wrapClass is not None:
            if pathOrObject is None:
                font = self.wrapClass()
            elif isinstance(pathOrObject, (str, os.PathLike)):
                font = self.wrapClass(os.fspath(pathOrObject))
            else:
                font = pathOrObject
            self._wrapped = font

    # path

    def _get_path(self, **kwargs: Any) -> str | None:
        return self.naked().path

    # save

    def _save(
        self,
        path: str | os.PathLike | None = None,
        showProgress: bool = False,
        formatVersion: int | None = None,
        fileStructure: str | None = None,
        **kwargs: Any,
    ) -> None:
        self.naked().save(
            path=path, formatVersion=formatVersion, structure=fileStructure
        )

    # close

    def _close(self, **kwargs: Any) -> None:
        del self._wrapped

    # -----------
    # Sub-Objects
    # -----------

    # info

    def _get_info(self) -> RInfo:
        return self.infoClass(pathOrObject=self.naked().info)

    # groups

    def _get_groups(self) -> RGroups:
        return self.groupsClass(pathOrObject=self.naked().groups)

    # kerning

    def _get_kerning(self) -> RKerning:
        return self.kerningClass(pathOrObject=self.naked().kerning)

    # features

    def _get_features(self) -> RFeatures:
        return self.featuresClass(pathOrObject=self.naked().features)

    # lib

    def _get_lib(self) -> RLib:
        return self.libClass(pathOrObject=self.naked().lib)

    # tempLib

    def _get_tempLib(self) -> RLib:
        return self.libClass(pathOrObject=self.naked().tempLib)

    # ------
    # Layers
    # ------

    def _get_layers(self, **kwargs: Any) -> tuple[RLayer, ...]:
        return tuple(
            self.layerClass(pathOrObject=layer) for layer in self.naked().layers
        )

    # order

    def _get_layerOrder(self, **kwargs: Any) -> tuple[str, ...]:
        return self.naked().layers.layerOrder

    def _set_layerOrder(self, value: CollectionType[str], **kwargs: Any) -> None:
        self.naked().layers.layerOrder = value

    # default layer

    def _get_defaultLayerName(self) -> str:
        return self.naked().layers.defaultLayer.name

    def _set_defaultLayerName(self, value: str, **kwargs: Any) -> None:
        font = self.naked()
        for layer in self.layers:
            if layer.name == value:
                break
        layer = layer.naked()
        font.layers.defaultLayer = layer

    # new

    def _newLayer(self, name: str, color: RGBALike | None, **kwargs: Any) -> RLayer:
        layers = self.naked().layers
        layer = layers.newLayer(name)
        layer.color = color
        return self.layerClass(pathOrObject=layer)

    # remove

    def _removeLayer(self, name: str, **kwargs: Any) -> None:
        layers = self.naked().layers
        del layers[name]

    # ------
    # Glyphs
    # ------

    def _get_glyphOrder(self) -> tuple[str, ...]:
        return self.naked().glyphOrder

    def _set_glyphOrder(self, value: CollectionType[str]) -> None:
        self.naked().glyphOrder = value

    # ----------
    # Guidelines
    # ----------

    def _lenGuidelines(self, **kwargs: Any) -> int:
        return len(self.naked().guidelines)

    def _getGuideline(self, index: int, **kwargs: Any) -> RGuideline:
        guideline = self.naked().guidelines[index]
        return self.guidelineClass(guideline)

    def _appendGuideline(
        self,
        position: CoordinateLike,
        angle: float | None,
        name: str | None = None,
        color: RGBALike | None = None,
        identifier: str | None = None,
        **kwargs: Any,
    ) -> RGuideline:
        font = self.naked()
        guideline = self.guidelineClass().naked()
        if guideline is None:
            raise ValueError("Guideline cannot be None.")
        guideline.x = position[0]
        guideline.y = position[1]
        guideline.angle = angle
        guideline.name = name
        guideline.color = color
        guideline.identifier = identifier
        font.appendGuideline(guideline)
        return self.guidelineClass(guideline)

    def _removeGuideline(self, index: int, **kwargs: Any) -> None:
        font = self.naked()
        guideline = font.guidelines[index]
        font.removeGuideline(guideline)



from __future__ import annotations
from collections.abc import ItemsView

import defcon
from fontParts.base import BaseGroups
from fontParts.base.annotations import CollectionType
from fontParts.fontshell.base import RBaseObject

ValueType = tuple[str, ...]
GroupsDict = dict[str, ValueType]


class RGroups(RBaseObject, BaseGroups):
    wrapClass = defcon.Groups

    def _get_side1KerningGroups(self) -> GroupsDict:
        groups = self.naked()
        representation = groups.getRepresentation("defcon.groups.kerningSide1Groups")
        return {k: tuple(v) for k, v in representation.items()}

    def _get_side2KerningGroups(self) -> GroupsDict:
        groups = self.naked()
        representation = groups.getRepresentation("defcon.groups.kerningSide2Groups")
        return {k: tuple(v) for k, v in representation.items()}

    def _items(self) -> ItemsView[str, tuple[str]]:
        groups = self.naked()
        formatted = {k: tuple(v) for k, v in groups.items()}
        return formatted.items()

    def _contains(self, key: str) -> bool:
        groups = self.naked()
        return key in groups

    def _setItem(self, key: str, value: CollectionType[str]) -> None:
        groups = self.naked()
        groups[key] = tuple(value)

    def _getItem(self, key: str) -> tuple[str]:
        groups = self.naked()
        return tuple(groups[key])

    def _delItem(self, key: str) -> None:
        groups = self.naked()
        del groups[key]



from __future__ import annotations

import defcon
from fontParts.base import BaseComponent
from fontParts.base.annotations import AffineTransformationLike, AffineTransformation
from fontParts.fontshell.base import RBaseObject


class RComponent(RBaseObject, BaseComponent):
    wrapClass: type[defcon.Component] = defcon.Component

    # ----------
    # Attributes
    # ----------

    # baseGlyph

    def _get_baseGlyph(self) -> str | None:
        component = self.naked()
        return component.baseGlyph if component else None

    def _set_baseGlyph(self, value: str) -> None:
        component = self.naked()
        if component is not None:
            component.baseGlyph = value

    # transformation

    def _get_transformation(self) -> AffineTransformation:
        component = self.naked()
        if component is None:
            raise ValueError("Component cannot be None.")
        return component.transformation

    def _set_transformation(self, value: AffineTransformationLike) -> None:
        component = self.naked()
        if component is not None:
            component.transformation = value

    # --------------
    # Identification
    # --------------

    # index

    def _set_index(self, value):
        component = self.naked()
        glyph = component.glyph
        if value > glyph.components.index(component):
            value -= 1
        glyph.removeComponent(component)
        glyph.insertComponent(value, component)

    # identifier

    def _get_identifier(self):
        component = self.naked()
        if component is None:
            return None
        return component.identifier

    def _getIdentifier(self):
        component = self.naked()
        if component is None:
            return None
        return component.generateIdentifier()

    def _setIdentifier(self, value):
        component = self.naked()
        if component is not None:
            component.identifier = value

    # -------------
    # Normalization
    # -------------

    def _decompose(self):
        component = self.naked()
        glyph = component.glyph
        glyph.decomposeComponent(component)



from __future__ import annotations
from typing import TYPE_CHECKING, Any
from collections.abc import Callable

from fontTools.misc import transform
from fontParts.base import normalizers
from fontParts.base.base import (
    BaseObject,
    TransformationMixin,
    InterpolationMixin,
    SelectionMixin,
    PointPositionMixin,
    IdentifierMixin,
    dynamicProperty,
    reference,
)
from fontParts.base.compatibility import AnchorCompatibilityReporter
from fontParts.base.color import Color
from fontParts.base.deprecated import DeprecatedAnchor, RemovedAnchor
from fontParts.base.annotations import (
    AffineTransformationLike,
    RGBALike,
    RGBA,
    IntFloatType,
)

if TYPE_CHECKING:
    from fontParts.base.font import BaseFont
    from fontParts.base.layer import BaseLayer
    from fontParts.base.glyph import BaseGlyph


class BaseAnchor(
    BaseObject,
    TransformationMixin,
    DeprecatedAnchor,
    RemovedAnchor,
    PointPositionMixin,
    InterpolationMixin,
    SelectionMixin,
    IdentifierMixin,
):
    """Represent the basis for an anchor object.

    This object is almost always created with :meth:`BaseGlyph.appendAnchor`.
    An orphan anchor can be created like this::

        >>> anchor = RAnchor()

    """

    def _reprContents(self) -> list[str]:
        contents = [f"({self.x}, {self.y})"]

        if self.name is not None:
            contents.append(f"name='{self.name}'")
        if self.color:
            contents.append(f"color={self.color!r}")
        return contents

    # ----
    # Copy
    # ----

    copyAttributes: tuple[str, ...] = ("x", "y", "name", "color")

    # -------
    # Parents
    # -------

    # Glyph

    _glyph: Callable[[], BaseGlyph] | None = None

    glyph: dynamicProperty = dynamicProperty(
        "glyph",
        """Get or set the anchor's parent glyph object.

        The value must be a :class:`BaseGlyph` instance or :obj:`None`.

        :return: The :class:`BaseGlyph` instance containing the anchor
            or :obj:`None`.
        :raises AssertionError: If attempting to set the glyph when it
            has already been set.

        Example::

            >>> glyph = anchor.glyph

        """,
    )

    def _get_glyph(self) -> BaseGlyph | None:
        if self._glyph is None:
            return None
        return self._glyph()

    def _set_glyph(self, glyph: BaseGlyph | Callable[[], BaseGlyph] | None) -> None:
        if self._glyph is not None:
            raise AssertionError("glyph for anchor already set")
        if glyph is not None:
            glyph = reference(glyph)
        self._glyph = glyph

    # Layer

    layer: dynamicProperty = dynamicProperty(
        "layer",
        """Get the anchor's parent layer object.

        This property is read-only.

        :return: The :class:`BaseLayer` instance containing the anchor
            or :obj:`None`.

        Example::

            >>> layer = anchor.layer

        """,
    )

    def _get_layer(self) -> BaseLayer | None:
        if self._glyph is None:
            return None
        return self.glyph.layer

    # Font

    font: dynamicProperty = dynamicProperty(
        "font",
        """Get the anchor's parent font object.

        This property is read-only.

        :return: The :class:`BaseFont` instance containing the anchor
            or :obj:`None`.

        Example::

            >>> font = anchor.font

        """,
    )

    def _get_font(self) -> BaseFont | None:
        if self._glyph is None:
            return None
        return self.glyph.font

    # --------
    # Position
    # --------

    # x

    x = dynamicProperty(
        "base_x",
        """Get or set the anchor's x-coordinate.

        The value must be an :class:`int` or a :class:`float`.

        :return: An :class:`int` or a :class:`float` representing the
            x-coordinate of the anchor.

        Example::

            >>> anchor.x
            100
            >>> anchor.x = 101

        """,
    )

    def _get_base_x(self) -> IntFloatType:
        value = self._get_x()
        value = normalizers.normalizeX(value)
        return value

    def _set_base_x(self, value: IntFloatType) -> None:
        value = normalizers.normalizeX(value)
        self._set_x(value)

    def _get_x(self) -> IntFloatType:
        """Get the native anchor's x-coordinate.

        This is the environment implementation of the :attr:`BaseAnchor.x` property
        getter.

        :return: An :class:`int` or a :class:`float` representing the
            x-coordinate of the anchor. The value will be normalized
            with :func:`normalizers.normalizeX`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_x(self, value: IntFloatType) -> None:
        """Set the native anchor's x-coordinate.

        This is the environment implementation of the :attr:`BaseAnchor.x` property
        setter.

        :param value: The x-coordinate to set as an :class:`int` or a :class:`float`.
            The value will have been normalized with :func:`normalizers.normalizeX`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # y

    y: dynamicProperty = dynamicProperty(
        "base_y",
        """Get or set the anchor's y-coordinate.

        The value must be an :class:`int` or a :class:`float`.

        :return: An :class:`int` or a :class:`float` representing the
            y-coordinate of the anchor.

        Example::

            >>> anchor.y
            100
            >>> anchor.y = 101

        """,
    )

    def _get_base_y(self) -> IntFloatType:
        value = self._get_y()
        value = normalizers.normalizeY(value)
        return value

    def _set_base_y(self, value: IntFloatType) -> None:
        value = normalizers.normalizeY(value)
        self._set_y(value)

    def _get_y(self) -> IntFloatType:
        """Get the native anchor's y-coordinate.

        This is the environment implementation of the :attr:`BaseAnchor.y` property
        getter.

        :return: An :class:`int` or a :class:`float` representing the
            x-coordinate of the anchor. The value will be normalized
            with :func:`normalizers.normalizeY`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_y(self, value: IntFloatType) -> None:
        """Set the native anchor's y-coordinate.

        This is the environment implementation of the :attr:`BaseAnchor.y` property
        setter.

        :param value: The y-coordinate to set as an :class:`int` or a :class:`float`.
            The value will have been normalized with :func:`normalizers.normalizeY`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # --------------
    # Identification
    # --------------

    # index

    index: dynamicProperty = dynamicProperty(
        "base_index",
        """Get the anchor's index.

        This property is read-only.

        :return: An :class:`int` representing the index of the anchor within
            the ordered list of the parent glyph's anchors.

        Example::

            >>> guideline.index
            0

        """,
    )

    def _get_base_index(self) -> int | None:
        value = self._get_index()
        value = normalizers.normalizeIndex(value)
        return value

    def _get_index(self) -> int | None:
        """Get the native anchor's index.

        This is the environment implementation of the :attr:`BaseAnchor.index`
        property getter.

        :return: An :class:`int` representing the index of the anchor within
            the ordered list of the parent glyph's anchors. The value will be
            normalized with :func:`normalizers.normalizeIndex`.

        .. note::

            Subclasses may override this method.

        """
        glyph = self.glyph
        if glyph is None:
            return None
        return glyph.anchors.index(self)

    # name

    name: dynamicProperty = dynamicProperty(
        "base_name",
        """Get or set the anchor's name.

        The value must be a :class:`str` or :obj:`None`.

        :return: A :class:`str` representing the name of the anchor, or :obj:`None`.

            >>> anchor.name
            'my anchor'
            >>> anchor.name = None

        """,
    )

    def _get_base_name(self) -> str | None:
        value = self._get_name()
        if value is not None:
            value = normalizers.normalizeAnchorName(value)
        return value

    def _set_base_name(self, value: str | None) -> None:
        if value is not None:
            value = normalizers.normalizeAnchorName(value)
        self._set_name(value)

    def _get_name(self) -> str | None:
        """Get the native anchor's name.

        This is the environment implementation of the :attr:`BaseAnchor.name`
        property getter.

        :return: A :class:`str` representing the name of the anchor,
            or :obj:`None`. The value will have been normalized
            with :func:`normalizers.normalizeAnchorName`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_name(self, value: str | None) -> None:
        """Set the native anchor's name.

        This is the environment implementation of the :attr:`BaseAnchor.name`
        property setter.

        :param value: The name to set as a :class:`str` or :obj:`None`. The
            value will have been normalized
            with :func:`normalizers.normalizeAnchorName`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # color

    color: dynamicProperty = dynamicProperty(
        "base_color",
        """Get or set the anchor's color.

        The value must be a :ref:`type-color` or :obj:`None`.

        :return: A :ref:`type-color` representing the color of the anchor,
            or :obj:`None`.

        Example::

            >>> anchor.color
            None
            >>> anchor.color = (1, 0, 0, 0.5)

        """,
    )

    def _get_base_color(self) -> Color | None:
        value = self._get_color()
        if value is not None:
            value = Color(value)
        return value

    def _set_base_color(self, value: RGBALike) -> None:
        if value is not None:
            value = normalizers.normalizeColor(value)
        self._set_color(value)

    def _get_color(self) -> RGBA | None:
        """Get the native anchor's color.

        This is the environment implementation of the :attr:`BaseAnchor.color`
        property getter.

        :return: A :ref:`type-color` representing the color of the anchor,
            or :obj:`None`. The value will be normalized
         with :func:`normalizers.normalizeColor`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_color(self, value: RGBALike | None) -> None:
        """Set the native anchor's color.

        Description

        This is the environment implementation of the :attr:`BaseAnchor.color`
        property setter.

        :param value: The :ref:`type-color` to set for the anchor or :obj:`None`.
            The value will have been normalized with :func:`normalizers.normalizeColor`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # --------------
    # Transformation
    # --------------

    def _transformBy(self, matrix: AffineTransformationLike, **kwargs: Any) -> None:
        r"""Transform the native anchor according to the given matrix.

        This is the environment implementation of :meth:`BaseAnchor.transformBy`.

        :param matrix: The :ref:`type-transformation` to apply. The value will
             be normalized with :func:`normalizers.normalizeTransformationMatrix`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        t = transform.Transform(*matrix)
        x, y = t.transformPoint((self.x, self.y))
        self.x = x
        self.y = y

    # -------------
    # Interpolation
    # -------------

    compatibilityReporterClass = AnchorCompatibilityReporter

    def isCompatible(
        self, other: BaseAnchor, cls=None
    ) -> tuple[bool, AnchorCompatibilityReporter]:
        """Evaluate interpolation compatibility with another anchor.

        :param other: The other :class:`BaseAnchor` instance to check
            compatibility with.
        :return: A :class:`tuple` where the first element is a :class:`bool`
            indicating compatibility, and the second element is
            a :class:`fontParts.base.compatibility.AnchorCompatibilityReporter`
            instance.

        Example::

            >>> compatible, report = self.isCompatible(otherAnchor)
            >>> compatible
            True
            >>> compatible
            [Warning] Anchor: "left" + "right"
            [Warning] Anchor: "left" has name left | "right" has name right

        """
        return super().isCompatible(other, BaseAnchor)

    def _isCompatible(
        self, other: BaseAnchor, reporter: AnchorCompatibilityReporter
    ) -> None:
        """Evaluate interpolation compatibility with another native anchor.

        This is the environment implementation of :meth:`BaseAnchor.isCompatible`.

        :param other: The other :class:`BaseAnchor` instance to check
            compatibility with.
        :param reporter: An object used to report compatibility issues.

        .. note::

            Subclasses may override this method.

        """
        anchor1 = self
        anchor2 = other
        # base names
        if anchor1.name != anchor2.name:
            reporter.nameDifference = True
            reporter.warning = True

    # -------------
    # Normalization
    # -------------

    def round(self) -> None:
        """Round the anchor's coordinate.

        This applies to:

        - :attr:`x`
        - :attr:`y

        Example::`

            >>> guideline.round()

        """
        self._round()

    def _round(self) -> None:
        """Round the native anchor's coordinate.

        This is the environment implementation of :meth:`BaseGuideline.round`.

        .. note::

            Subclasses may override this method.

        """
        self.x = normalizers.normalizeVisualRounding(self.x)
        self.y = normalizers.normalizeVisualRounding(self.y)
